from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_ITERATE_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession

import transaction
import unittest


class TestWorkingCopyEndpoint(unittest.TestCase):

    layer = PLONE_RESTAPI_ITERATE_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        self.portal.invokeFactory("Document", "document", title="Test Document")
        self.doc = self.portal.document

        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_workingcopy_checkout(self):
        response = self.api_session.post(
            "/document/@workingcopy",
        )

        self.assertEqual(response.status_code, 201)
        self.assertIn("@id", response.json())

        self.assertEqual(
            response.json()["@id"],
            f"{self.portal_url}/copy_of_document",
        )

    def test_workingcopy_get(self):
        # We create the working copy
        response = self.api_session.post(
            "/document/@workingcopy",
        )

        self.assertEqual(response.status_code, 201)

        # endpoint GET in the baseline object
        response = self.api_session.get(
            "/document/@workingcopy",
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json()["working_copy"]["@id"],
            f"{self.portal_url}/copy_of_document",
        )
        self.assertEqual(
            response.json()["working_copy"]["creator_name"],
            "admin",
        )
        self.assertEqual(
            response.json()["working_copy"]["creator_url"],
            f"{self.portal_url}/author/admin",
        )

        # endpoint GET in the working_copy
        response = self.api_session.get(
            "/copy_of_document/@workingcopy",
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json()["working_copy_of"]["@id"],
            f"{self.portal_url}/document",
        )
        self.assertEqual(
            response.json()["working_copy"]["@id"],
            f"{self.portal_url}/copy_of_document",
        )
        self.assertEqual(
            response.json()["working_copy"]["creator_name"],
            "admin",
        )
        self.assertEqual(
            response.json()["working_copy"]["creator_url"],
            f"{self.portal_url}/author/admin",
        )

        # Serialization on the baseline object
        response = self.api_session.get(
            "/document",
        )

        self.assertEqual(
            response.json()["working_copy"]["@id"],
            f"{self.portal_url}/copy_of_document",
        )
        self.assertEqual(
            response.json()["working_copy"]["creator_name"],
            "admin",
        )
        self.assertEqual(
            response.json()["working_copy"]["creator_url"],
            f"{self.portal_url}/author/admin",
        )
        self.assertEqual(response.json()["working_copy_of"], None)

        # Serialization on the working copy object
        response = self.api_session.get(
            "/copy_of_document",
        )
        self.assertEqual(
            response.json()["working_copy_of"]["@id"],
            f"{self.portal_url}/document",
        )
        self.assertEqual(
            response.json()["working_copy"]["@id"],
            f"{self.portal_url}/copy_of_document",
        )
        self.assertEqual(
            response.json()["working_copy"]["creator_name"],
            "admin",
        )
        self.assertEqual(
            response.json()["working_copy"]["creator_url"],
            f"{self.portal_url}/author/admin",
        )

    def test_workingcopy_notworkingcopy_get(self):
        # endpoint GET in the working_copy
        response = self.api_session.get(
            "/document/",
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            response.json()["working_copy_of"],
            None,
        )

    def test_workingcopy_delete_on_the_baseline(self):
        # We create the working copy
        response = self.api_session.post(
            "/document/@workingcopy",
        )
        self.assertEqual(response.status_code, 201)

        # Deleting in the baseline deletes the working copy
        response = self.api_session.delete(
            "/document/@workingcopy",
        )

        self.assertEqual(response.status_code, 204)

        response = self.api_session.get(
            "/copy_of_document",
        )
        self.assertEqual(response.status_code, 404)

    def test_workingcopy_delete_on_the_working_copy(self):
        # We create the working copy
        response = self.api_session.post(
            "/document/@workingcopy",
        )
        self.assertEqual(response.status_code, 201)

        # Deleting in the working copy deletes it too
        response = self.api_session.delete(
            "/copy_of_document/@workingcopy",
        )

        self.assertEqual(response.status_code, 204)

        response = self.api_session.get(
            "/copy_of_document",
        )
        self.assertEqual(response.status_code, 404)

    def test_workingcopy_patch_on_the_baseline(self):
        # We create the working copy
        response = self.api_session.post(
            "/document/@workingcopy",
        )
        self.assertEqual(response.status_code, 201)

        # Modify the working copy
        response = self.api_session.patch(
            "/copy_of_document", json={"title": "I just changed the title"}
        )

        # Checking in
        response = self.api_session.patch(
            "/document/@workingcopy",
        )

        # Check if the change is there
        response = self.api_session.get(
            "/document",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["title"], "I just changed the title")

    def test_workingcopy_patch_on_the_working_copy(self):
        # We create the working copy
        response = self.api_session.post(
            "/document/@workingcopy",
        )
        self.assertEqual(response.status_code, 201)

        # Modify the working copy
        response = self.api_session.patch(
            "/copy_of_document", json={"title": "I just changed the title"}
        )

        # Checking in
        response = self.api_session.patch(
            "/copy_of_document/@workingcopy",
        )

        # Check if the change is there
        response = self.api_session.get(
            "/document",
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["title"], "I just changed the title")



from datetime import datetime
from datetime import time
from datetime import timedelta
from plone.app.textfield import RichText
from plone.app.vocabularies.catalog import CatalogSource
from plone.autoform.directives import read_permission
from plone.autoform.directives import write_permission
from plone.autoform.interfaces import IFormFieldProvider
from plone.dexterity.content import Item
from plone.namedfile import field as namedfile
from plone.restapi.tests.helpers import ascii_token
from plone.supermodel import model
from plone.supermodel.directives import primary
from Products.CMFCore.utils import getToolByName
from pytz import timezone
from z3c.formwidget.query.interfaces import IQuerySource
from z3c.relationfield.schema import RelationChoice
from z3c.relationfield.schema import RelationList
from zope import schema
from zope.interface import directlyProvides
from zope.interface import implementer
from zope.interface import Invalid
from zope.interface import invariant
from zope.interface import provider
from zope.schema.interfaces import IContextAwareDefaultFactory
from zope.schema.interfaces import IContextSourceBinder
from zope.schema.interfaces import IIterableSource
from zope.schema.interfaces import ISource
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


INDEXES = (
    ("test_int_field", "FieldIndex"),
    ("test_list_field", "KeywordIndex"),
    ("test_bool_field", "BooleanIndex"),
)


@implementer(ISource)
class MyNonIterableSource:
    divisor = 2

    def __contains__(self, value):
        return bool(value % self.divisor)


@implementer(IIterableSource)
class MyIterableSource:
    values = [1, 2, 3]

    def __contains__(self, value):
        return value in self.values

    def __iter__(self):
        terms = [
            SimpleTerm(value=v, token="token%s" % v, title="Title %s" % v)
            for v in self.values
        ]
        return iter(terms)


@implementer(IQuerySource)
class MyIterableQuerySource:
    values = [1, 2, 3]

    def __contains__(self, value):
        return value in self.values

    def search(self, query):
        terms = [
            SimpleTerm(value=v, token="token%s" % v, title="Title %s" % v)
            for v in self.values
        ]
        return [t for t in terms if query in str(t.token)]

    def __iter__(self):
        # The @querysources endpoint should never attempt to enumerate terms
        raise NotImplementedError


@implementer(IIterableSource)
class MyIterableContextSource:
    def __init__(self, context):
        self.context = context

        title_words = self.context.title.split()
        self.terms = [
            SimpleTerm(value=w.lower(), token=ascii_token(w.lower()), title=w)
            for w in title_words
        ]

    def __contains__(self, value):
        return value in [t.value for t in self.terms]

    def __iter__(self):
        return iter(self.terms)


@implementer(IQuerySource)
class MyContextQuerySource:
    def __init__(self, context):
        self.context = context

        title_words = self.context.title.split()
        self.terms = [
            SimpleTerm(value=w.lower(), token=ascii_token(w.lower()), title=w)
            for w in title_words
        ]

    def __contains__(self, value):
        return value in [t.value for t in self.terms]

    def __iter__(self):
        # The @querysources endpoint should never attempt to enumerate terms
        raise NotImplementedError

    def search(self, query):
        return [t for t in iter(self.terms) if query in str(t.token)]


my_iterable_source = MyIterableSource()
my_non_iterable_source = MyNonIterableSource()
my_querysource = MyIterableQuerySource()


def my_context_source_binder(context):
    return MyIterableContextSource(context)


directlyProvides(my_context_source_binder, IContextSourceBinder)


def my_context_querysource_binder(context):
    return MyContextQuerySource(context)


directlyProvides(my_context_querysource_binder, IContextSourceBinder)


def vocabularyRequireingContextFactory(context):
    catalog = getToolByName(context, "portal_catalog")
    return SimpleVocabulary([SimpleTerm(catalog.id, catalog.id, catalog.id)])


class IDXTestDocumentSchema(model.Schema):

    # zope.schema fields
    test_ascii_field = schema.ASCII(required=False)
    test_asciiline_field = schema.ASCIILine(required=False)
    test_bool_field = schema.Bool(required=False)
    test_bytes_field = schema.Bytes(required=False)
    test_bytesline_field = schema.BytesLine(required=False)
    test_choice_field = schema.Choice(values=["foo", "bar"], required=False)
    test_choice_field_with_vocabulary = schema.Choice(
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm("value1", "token1", "title1"),
                SimpleTerm("value2", "token2", "title2"),
            ]
        ),
        required=False,
    )

    test_choice_with_non_iterable_source = schema.Choice(
        required=False, source=my_non_iterable_source
    )
    test_choice_with_source = schema.Choice(required=False, source=my_iterable_source)
    test_choice_with_context_source = schema.Choice(
        required=False, source=my_context_source_binder
    )
    test_choice_with_querysource = schema.Choice(required=False, source=my_querysource)
    test_choice_with_context_querysource = schema.Choice(
        required=False, source=my_context_querysource_binder
    )

    test_date_field = schema.Date(required=False)
    test_datetime_field = schema.Datetime(required=False)
    test_datetime_tz_field = schema.Datetime(
        required=False,
        defaultFactory=lambda: timezone("Europe/Zurich").localize(
            datetime(2017, 10, 31, 10, 0)
        ),
    )
    test_decimal_field = schema.Decimal(required=False)
    test_dict_field = schema.Dict(required=False)
    test_float_field = schema.Float(required=False)
    test_frozenset_field = schema.FrozenSet(required=False)
    test_int_field = schema.Int(required=False)
    test_list_field = schema.List(required=False)
    test_list_field_with_choice_with_vocabulary = schema.List(
        value_type=schema.Choice(
            vocabulary=SimpleVocabulary(
                [
                    SimpleTerm("value1", "token1", "title1"),
                    SimpleTerm("value2", "token2", "title2"),
                    SimpleTerm("value3", "token3", "title3"),
                ]
            )
        ),
        required=False,
    )
    test_set_field = schema.Set(required=False)
    test_set_field_with_choice_with_vocabulary = schema.Set(
        value_type=schema.Choice(
            vocabulary=SimpleVocabulary(
                [
                    SimpleTerm("value1", "token1", "title1"),
                    SimpleTerm("value2", "token2", "title2"),
                    SimpleTerm("value3", "token3", "title3"),
                ]
            )
        ),
        required=False,
    )
    test_text_field = schema.Text(required=False)
    test_textline_field = schema.TextLine(required=False)
    test_time_field = schema.Time(required=False)
    test_timedelta_field = schema.Timedelta(required=False)
    test_tuple_field = schema.Tuple(required=False)
    test_nested_list_field = schema.List(required=False, value_type=schema.Tuple())
    test_nested_dict_field = schema.Dict(
        required=False, key_type=schema.ASCIILine(), value_type=schema.Tuple()
    )
    test_list_choice_with_context_vocabulary_field = schema.List(
        title="Field",
        value_type=schema.Choice(vocabulary="plone.restapi.testing.context_vocabulary"),
        required=False,
    )

    # plone.app.textfield
    test_richtext_field = RichText(
        required=False, allowed_mime_types=["text/html", "text/plain"]
    )

    # plone.namedfile fields
    test_namedfile_field = namedfile.NamedFile(required=False)
    test_namedimage_field = namedfile.NamedImage(required=False)
    test_namedblobfile_field = namedfile.NamedBlobFile(required=False)
    test_namedblobimage_field = namedfile.NamedBlobImage(required=False)

    primary("test_primary_namedfile_field")
    test_primary_namedfile_field = namedfile.NamedFile(required=False)

    # z3c.relationfield
    test_relationchoice_field = RelationChoice(
        required=False, source=CatalogSource(id=["doc1", "doc2"])
    )
    test_relationlist_field = RelationList(
        required=False,
        value_type=RelationChoice(vocabulary="plone.app.vocabularies.Catalog"),
    )

    # Test fields for validation
    test_required_field = schema.TextLine(required=True)
    test_readonly_field = schema.TextLine(required=False, readonly=True)
    test_maxlength_field = schema.TextLine(required=False, max_length=10)
    test_constraint_field = schema.TextLine(
        required=False, constraint=lambda x: "00" in x
    )
    test_datetime_min_field = schema.Datetime(required=False, min=datetime(2000, 1, 1))
    test_time_min_field = schema.Time(required=False, min=time(1))
    test_timedelta_min_field = schema.Timedelta(required=False, min=timedelta(100))
    test_list_value_type_field = schema.List(required=False, value_type=schema.Int())
    test_dict_key_type_field = schema.Dict(required=False, key_type=schema.Int())

    read_permission(test_read_permission_field="cmf.ManagePortal")
    test_read_permission_field = schema.TextLine(required=False)
    write_permission(test_write_permission_field="cmf.ManagePortal")
    test_write_permission_field = schema.TextLine(required=False)

    read_permission(test_read_permission_field="cmf.ManagePortal")
    test_read_permission_field = schema.TextLine(required=False)

    test_invariant_field1 = schema.TextLine(required=False)
    test_invariant_field2 = schema.TextLine(required=False)

    test_missing_value_field = schema.TextLine(
        required=False, missing_value="missing", default="default"
    )

    test_missing_value_required_field = schema.TextLine(
        required=True, missing_value="missing", default="some value"
    )

    @invariant
    def validate_same_value(data):
        if data.test_invariant_field1 != data.test_invariant_field2:
            raise Invalid("Must have same values")

    # Test fields with default values
    test_default_value_field = schema.TextLine(required=True, default="Default")

    @provider(IContextAwareDefaultFactory)
    def default_factory(context):
        return "DefaultFactory"

    test_default_factory_field = schema.TextLine(
        required=True, defaultFactory=default_factory
    )


class DXTestDocument(Item):
    """A Dexterity based test type containing a set of standard fields."""

    # Plone standard types (both, dx and at) do provide exclude_from_nav
    # and at least one test expect it to be here, so make it explicit
    exclude_from_nav = False


@provider(IFormFieldProvider)
class ITestBehavior(model.Schema):

    test_behavior_field = schema.TextLine(required=False)
    # Add nav_title to test if it gets substituted in Navigation service
    nav_title = schema.TextLine(required=False)


@provider(IFormFieldProvider)
class ITestAnnotationsBehavior(model.Schema):

    test_annotations_behavior_field = schema.TextLine(required=False)



from functools import reduce


def unflatten_dotted_dict(dct):
    """Reconstructs nesting for a dict that has been flattened to one level.

    Further nesting is indicated by dots in the key names: Keys will be merged
    together into sub-dicts based on their longest common prefix.

    Example:

    >>> dct = {
    ...     'a.b.X': 1,
    ...     'a.b.Y': 2,
    ...     'a.foo': 3,
    ...     'bar': 4,
    ... }
    >>>
    >>> unflatten_dotted_dict(dct)

    {'a': {'b': {'X': 1,
                 'Y': 2},
           'foo': 3},
     'bar': 4}
    """

    def create_or_get(dct, key):
        return dct.setdefault(key, {})

    result = {}

    for key, value in dct.items():
        key_segments = key.split(".")
        # Create nested dicts from parent keys, if any
        inner = reduce(create_or_get, [result] + key_segments[:-1])
        # Assign value to terminal key
        inner[key_segments[-1]] = value

    return result



from plone.locking.interfaces import INonStealableLock
from plone.locking.interfaces import IRefreshableLockable
from plone.restapi.deserializer import json_body
from plone.restapi.services import Service
from plone.restapi.services.locking import lock_info
from plone.restapi.services.locking import webdav_lock
from zope.interface import alsoProvides

import plone.protect.interfaces


class Lock(Service):
    """Lock an object"""

    def reply(self):
        data = json_body(self.request)

        lockable = IRefreshableLockable(self.context, None)
        if lockable is not None:
            lockable.lock()

            if "stealable" in data and not data["stealable"]:
                alsoProvides(self.context, INonStealableLock)

            if "timeout" in data:
                lock_item = webdav_lock(self.context)
                lock_item.setTimeout("Second-%s" % data["timeout"])

            # Disable CSRF protection
            if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
                alsoProvides(
                    self.request, plone.protect.interfaces.IDisableCSRFProtection
                )

        return lock_info(self.context)



from plone.restapi.batching import HypermediaBatch
from plone.restapi.interfaces import ISerializeToJson
from Products.CMFPlone.utils import safe_unicode
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.i18n import translate
from zope.interface import implementer
from zope.interface import Interface
from zope.schema.interfaces import IIterableSource
from zope.schema.interfaces import ITitledTokenizedTerm
from zope.schema.interfaces import ITokenizedTerm
from zope.schema.interfaces import IVocabulary

import warnings


@implementer(ISerializeToJson)
class SerializeVocabLikeToJson:
    """Base implementation to serialize vocabularies and sources to JSON.

    Implements server-side filtering as well as batching.
    """

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, vocabulary_id):
        vocabulary = self.context
        title = safe_unicode(self.request.form.get("title", ""))
        token = self.request.form.get("token", "")
        tokens = self.request.form.get("tokens", [])
        b_size = self.request.form.get("b_size", "")

        terms = []
        for term in vocabulary:
            if title and token:
                self.request.response.setStatus(400)
                return dict(
                    error=dict(
                        type="Invalid parameters",
                        message="You can not filter by title and token at the same time.",
                    )  # noqa
                )

            if token:
                warnings.warn(
                    "``token`` parameter is deprecated and will be removed in plone.restapi 9.0. Use ``tokens`` parameter instead.",
                    DeprecationWarning,
                )
                if token.lower() != term.token.lower():
                    continue
                terms.append(term)
            elif tokens:
                if isinstance(tokens, str):
                    tokens = [tokens]
                for item in tokens:
                    if item.lower() != term.token.lower():
                        continue
                    terms.append(term)
            else:
                term_title = safe_unicode(getattr(term, "title", None) or "")
                if (
                    title.lower()
                    not in translate(term_title, context=self.request).lower()
                ):
                    continue
                terms.append(term)

        serialized_terms = []

        # Do not batch parameter is set
        if b_size == "-1":
            for term in terms:
                serializer = getMultiAdapter(
                    (term, self.request), interface=ISerializeToJson
                )
                serialized_terms.append(serializer())
            return {"@id": vocabulary_id, "items": serialized_terms}

        batch = HypermediaBatch(self.request, terms)

        for term in batch:
            serializer = getMultiAdapter(
                (term, self.request), interface=ISerializeToJson
            )
            serialized_terms.append(serializer())

        result = {
            "@id": batch.canonical_url,
            "items": serialized_terms,
            "items_total": batch.items_total,
        }
        links = batch.links
        if links:
            result["batching"] = links
        return result


@adapter(IVocabulary, Interface)
class SerializeVocabularyToJson(SerializeVocabLikeToJson):
    """Serializes IVocabulary to JSON."""


@adapter(IIterableSource, Interface)
class SerializeSourceToJson(SerializeVocabLikeToJson):
    """Serializes IIterableSource to JSON."""


@implementer(ISerializeToJson)
@adapter(ITokenizedTerm, Interface)
class SerializeTermToJson:
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self):
        term = self.context
        token = term.token
        title = term.title if ITitledTokenizedTerm.providedBy(term) else token
        if isinstance(title, bytes):
            title = title.decode("UTF-8")
        return {"token": token, "title": translate(title, context=self.request)}



from plone import api
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession

import transaction
import unittest


class TestHistoryEndpoint(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        self.portal.invokeFactory(
            "Document", id="doc_with_history", title="My Document"
        )
        self.doc = self.portal.doc_with_history
        self.doc.setTitle("Current version")

        api.content.transition(self.doc, "publish")

        self.endpoint_url = f"{self.doc.absolute_url()}/@history"

        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_get_types(self):
        # Check if we have all history types in our test setup
        response = self.api_session.get(self.endpoint_url)
        data = response.json()

        types = [item["type"] for item in data]

        self.assertEqual({"versioning", "workflow"}, set(types))

    def test_get_datastructure(self):
        response = self.api_session.get(self.endpoint_url)
        data = response.json()

        actor_keys = ["@id", "id", "fullname", "username"]

        main_keys = ["action", "actor", "comments", "time", "transition_title", "type"]

        history_keys = main_keys + ["@id", "may_revert", "version"]

        workflow_keys = main_keys + ["review_state", "state_title"]

        for item in data:
            # Make sure we'll add tests when new history types are added.
            self.assertIn(item["type"], ["versioning", "workflow"])

            if item["type"] == "versioning":
                self.assertEqual(set(item), set(history_keys))
            else:
                self.assertEqual(set(item), set(workflow_keys))

            self.assertEqual(set(item["actor"]), set(actor_keys))

            self.assertIsNotNone(item["action"])

    def test_revert(self):
        url = f"{self.doc.absolute_url()}/@history"
        response = self.api_session.patch(url, json={"version": 0})
        self.assertEqual(response.status_code, 200)

        # My Document is the old title
        self.assertEqual(
            response.json(),
            {"message": "My Document has been reverted to revision 0."},
        )

    def test_time_field(self):
        url = f"{self.doc.absolute_url()}/@history"
        response = self.api_session.get(url)

        for item in response.json():
            self.assertTrue(isinstance(item["time"], str))

    def test_get_historical_link(self):
        # The @id field should link to @history/version.
        response = self.api_session.get(self.endpoint_url)
        data = response.json()

        for item in data:
            if item["type"] == "versioning":
                self.assertTrue(
                    item["@id"].endswith("@history/" + str(item["version"]))
                )
            else:
                self.assertNotIn("@id", list(item))

    def test_explicit_current(self):
        # Does version=current get the current version
        url = self.doc.absolute_url() + "/@history/current"
        response = self.api_session.get(url)
        self.assertEqual(response.json()["title"], "Current version")

    def test_previous_version(self):
        # Does version=0 get the older version?
        url = self.doc.absolute_url() + "/@history/0"
        response = self.api_session.get(url)
        self.assertEqual(response.json()["title"], "My Document")

    def test_no_sharing(self):
        url = self.doc.absolute_url() + "/@history/0"
        response = self.api_session.get(url)
        self.assertNotIn("sharing", response.json())


class TestHistoryEndpointEmptyOrInacessibleHistory(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def _disable_auto_versioning(self, content_type):
        portal_repository = self.portal.portal_repository
        types = list(portal_repository.getVersionableContentTypes())
        types.remove(content_type)
        portal_repository.setVersionableContentTypes(types)
        portal_repository.removePolicyFromContentType(content_type, "version_on_revert")

    def setUp(self):
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()

        login(self.portal, SITE_OWNER_NAME)

        # disabling auto versioning is necessary to have an empty revision
        # history
        self._disable_auto_versioning("Document")

        self.portal.invokeFactory(
            "Document", id="doc_with_empty_history", title="My Document"
        )
        self.doc = self.portal.doc_with_empty_history
        api.content.transition(self.doc, "publish")
        self.endpoint_url = f"{self.doc.absolute_url()}/@history"

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (TEST_USER_NAME, TEST_USER_PASSWORD)
        # forbid access to `workflowHistory`
        setRoles(self.portal, TEST_USER_ID, ["Reader", "Contributor"])
        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_empty_or_inaccessible_full_history_returns_empty_list(self):
        url = self.doc.absolute_url() + "/@history"
        response = self.api_session.get(url)
        self.assertEqual([], response.json())


class TestHistoryEndpointTranslatedMessages(unittest.TestCase):
    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.headers.update({"Accept-Language": "es"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        self.portal.invokeFactory(
            "Document", id="doc_with_history", title="My Document"
        )
        self.doc = self.portal.doc_with_history
        self.doc.setTitle("Current version")

        api.content.transition(self.doc, "publish")

        self.endpoint_url = f"{self.doc.absolute_url()}/@history"

        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_actions_are_translated(self):
        url = self.doc.absolute_url() + "/@history"
        response = self.api_session.get(url)
        first_action = response.json()[-1]
        self.assertEqual("Crear", first_action["action"])

    def test_state_titles_are_translated(self):
        url = self.doc.absolute_url() + "/@history"
        response = self.api_session.get(url)
        first_action = response.json()[-1]
        self.assertEqual("Privado", first_action["state_title"])

    def test_transition_titles_are_translated(self):
        url = self.doc.absolute_url() + "/@history"
        response = self.api_session.get(url)
        first_action = response.json()[-1]
        self.assertEqual("Crear", first_action["transition_title"])



from plone.locking.interfaces import IRefreshableLockable
from plone.restapi.services import Service
from plone.restapi.services.locking import lock_info
from zope.interface import alsoProvides

import plone.protect.interfaces


class Lock(Service):
    """Refresh the lock of an object"""

    def reply(self):
        lockable = IRefreshableLockable(self.context, None)
        if lockable is not None:
            lockable.refresh_lock()

            # Disable CSRF protection
            if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
                alsoProvides(
                    self.request, plone.protect.interfaces.IDisableCSRFProtection
                )

        return lock_info(self.context)



from plone.registry.interfaces import IRegistry
from plone.restapi.services import Service
from zope.component import getUtility
from zope.interface import alsoProvides

import json
import plone.protect.interfaces


class RegistryUpdate(Service):
    def reply(self):
        records_to_update = json.loads(self.request.get("BODY", "{}"))
        registry = getUtility(IRegistry)

        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        for key, value in records_to_update.items():
            if key not in registry:
                raise NotImplementedError(
                    "This endpoint is only intended to update existing "
                    "records! Couldn't find key %r" % key
                )
            registry[key] = value
        return self.reply_no_content()



from plone.dexterity.interfaces import IDexterityContent
from plone.restapi.deserializer.dxfields import DefaultFieldDeserializer
from plone.restapi.interfaces import IFieldDeserializer
from Products.CMFCore.utils import getToolByName
from z3c.relationfield.interfaces import IRelationChoice
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.component import queryUtility
from zope.interface import implementer
from zope.intid.interfaces import IIntIds
from zope.publisher.interfaces.browser import IBrowserRequest


@implementer(IFieldDeserializer)
@adapter(IRelationChoice, IDexterityContent, IBrowserRequest)
class RelationChoiceFieldDeserializer(DefaultFieldDeserializer):
    def __call__(self, value):
        obj = None

        if isinstance(value, dict):
            # We are trying to deserialize the output of a serialization
            # which is enhanced, extract it and put it on the loop again
            value = value["@id"]

        if isinstance(value, int):
            # Resolve by intid
            intids = queryUtility(IIntIds)
            obj = intids.queryObject(value)
            resolved_by = "intid"
        elif isinstance(value, str):
            portal = getMultiAdapter(
                (self.context, self.request), name="plone_portal_state"
            ).portal()
            portal_url = portal.absolute_url()
            if value.startswith(portal_url):
                # Resolve by URL
                obj = portal.restrictedTraverse(value[len(portal_url) + 1 :], None)
                resolved_by = "URL"
            elif value.startswith("/"):
                # Resolve by path
                obj = portal.restrictedTraverse(value.lstrip("/"), None)
                resolved_by = "path"
            else:
                # Resolve by UID
                catalog = getToolByName(self.context, "portal_catalog")
                brain = catalog(UID=value)
                if brain:
                    obj = brain[0].getObject()
                resolved_by = "UID"

        if obj is None:
            self.request.response.setStatus(400)
            raise ValueError(f"Could not resolve object for {resolved_by}={value}")

        self.field.validate(obj)
        return obj



from DateTime import DateTime
from importlib import import_module
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.textfield.value import RichTextValue
from plone.namedfile.file import NamedBlobImage
from plone.namedfile.file import NamedFile
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from plone.scale import storage
from Products.CMFCore.utils import getToolByName
from unittest.mock import patch
from zope.component import getMultiAdapter

import json
import os
import unittest


HAS_PLONE_6 = getattr(
    import_module("Products.CMFPlone.factory"), "PLONE60MARKER", False
)


class TestSerializeToJsonAdapter(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.workflowTool = getToolByName(self.portal, "portal_workflow")
        self.portal_url = self.portal.absolute_url()
        self.portal.invokeFactory("Document", id="doc1", title="Document 1")
        self.portal.invokeFactory(
            "DXTestDocument", id="dxdoc", title="DX Test Document"
        )

    def serialize(self, obj, fullobjects=False):
        if fullobjects:
            self.request.form["fullobjects"] = 1

        serializer = getMultiAdapter((obj, self.request), ISerializeToJson)
        return serializer()

    def test_serialize_returns_id(self):
        self.assertEqual(
            self.serialize(self.portal.doc1)["@id"], self.portal_url + "/doc1"
        )

    def test_serialize_returns_type(self):
        self.assertTrue(
            self.serialize(self.portal.doc1).get("@type"),
            "The @type attribute should be present.",
        )
        self.assertEqual(self.serialize(self.portal.doc1)["@type"], "Document")

    def test_serialize_returns_title(self):
        self.assertEqual(self.serialize(self.portal.doc1)["title"], "Document 1")

    def test_serialize_can_read_as_manager(self):
        self.portal.dxdoc.test_read_permission_field = "Test Read Permission"
        self.workflowTool.doActionFor(self.portal.dxdoc, "publish")
        setRoles(self.portal, TEST_USER_ID, ["Member", "Manager"])
        self.assertIn(
            "Test Read Permission", list(self.serialize(self.portal.dxdoc).values())
        )

    def test_serialize_cannot_read_as_member(self):
        self.portal.dxdoc.test_read_permission_field = "Test Read Permission"
        self.workflowTool.doActionFor(self.portal.dxdoc, "publish")
        setRoles(self.portal, TEST_USER_ID, ["Member"])
        self.assertNotIn(
            "Test Read Permission", list(self.serialize(self.portal.dxdoc).values())
        )

    def test_serialize_returns_desciption(self):
        self.portal.doc1.description = "This is a document"
        self.assertEqual(
            self.serialize(self.portal.doc1)["description"], "This is a document"
        )

    def test_serialize_returns_rich_text(self):
        self.portal.doc1.text = RichTextValue("Lorem ipsum.", "text/plain", "text/html")
        self.assertEqual(
            self.serialize(self.portal.doc1).get("text"),
            {
                "data": "<p>Lorem ipsum.</p>",
                "content-type": "text/plain",
                "encoding": "utf-8",
            },
        )

    def test_serialize_returns_effective(self):
        self.portal.doc1.setEffectiveDate(DateTime("2014/04/04"))
        self.assertEqual(
            self.serialize(self.portal.doc1)["effective"], "2014-04-04T00:00:00"
        )

    def test_serialize_returns_expires(self):
        self.portal.doc1.setExpirationDate(DateTime("2017/01/01"))
        self.assertEqual(
            self.serialize(self.portal.doc1)["expires"], "2017-01-01T00:00:00"
        )

    def test_serialize_on_folder_returns_items_attr(self):
        self.portal.invokeFactory("Folder", id="folder1", title="Folder 1")
        self.portal.folder1.invokeFactory("Document", id="doc1")
        self.portal.folder1.doc1.title = "Document 1"
        self.portal.folder1.doc1.description = "This is a document"
        self.portal.folder1.doc1.reindexObject()
        self.assertEqual(
            self.serialize(self.portal.folder1)["items"],
            [
                {
                    "@id": "http://nohost/plone/folder1/doc1",
                    "@type": "Document",
                    "description": "This is a document",
                    "title": "Document 1",
                    "review_state": "private",
                }
            ],
        )

    def test_serialize_folder_orders_items_by_get_object_position_in_parent(
        self,
    ):  # noqa
        self.portal.invokeFactory("Folder", id="folder1", title="Folder 1")
        self.portal.folder1.invokeFactory("Document", id="doc1")
        self.portal.folder1.doc1.title = "Document 1"
        self.portal.folder1.doc1.description = "This is a document"
        self.portal.folder1.doc1.reindexObject()

        self.portal.folder1.invokeFactory("Document", id="doc2")
        self.portal.folder1.doc2.title = "Document 2"
        self.portal.folder1.doc2.description = "Second doc"
        self.portal.folder1.doc2.reindexObject()

        # Change GOPIP (getObjectPositionInParent) based order
        self.portal.folder1.moveObjectsUp("doc2")

        self.assertEqual(
            self.serialize(self.portal.folder1)["items"],
            [
                {
                    "@id": "http://nohost/plone/folder1/doc2",
                    "@type": "Document",
                    "description": "Second doc",
                    "title": "Document 2",
                    "review_state": "private",
                },
                {
                    "@id": "http://nohost/plone/folder1/doc1",
                    "@type": "Document",
                    "description": "This is a document",
                    "title": "Document 1",
                    "review_state": "private",
                },
            ],
        )

    def test_serialize_returns_parent(self):
        self.assertTrue(
            self.serialize(self.portal.doc1).get("parent"),
            "The parent attribute should be present.",
        )
        self.assertEqual(
            {
                "@id": self.portal.absolute_url(),
                "@type": self.portal.portal_type,
                "title": self.portal.title,
                "description": self.portal.description,
            },
            self.serialize(self.portal.doc1)["parent"],
        )

    def test_serialize_does_not_returns_parent_on_root(self):
        self.assertEqual(
            {},
            self.serialize(self.portal).get("parent"),
            "The parent attribute should be present, even on portal root.",
        )
        self.assertEqual(
            {
                "@id": self.portal.absolute_url(),
                "@type": self.portal.portal_type,
                "title": self.portal.title,
                "description": self.portal.description,
            },
            self.serialize(self.portal.doc1)["parent"],
            "The parent attribute on portal root should be None",
        )

    def test_serialize_returns_site_root_type(self):
        self.assertTrue(
            self.serialize(self.portal).get("@type"),
            "The @type attribute should be present.",
        )
        self.assertEqual(self.serialize(self.portal)["@type"], "Plone Site")

    def test_serialize_site_orders_items_by_get_object_position_in_parent(self):  # noqa
        # Change GOPIP (getObjectPositionInParent) based order
        self.portal.moveObjectsUp("dxdoc")

        self.assertEqual(
            self.serialize(self.portal)["items"],
            [
                {
                    "@id": "http://nohost/plone/dxdoc",
                    "@type": "DXTestDocument",
                    "description": "",
                    "title": "DX Test Document",
                    "review_state": "private",
                },
                {
                    "@id": "http://nohost/plone/doc1",
                    "@type": "Document",
                    "description": "",
                    "title": "Document 1",
                    "review_state": "private",
                },
            ],
        )

    def test_serialize_ignores_underscore_values(self):
        self.assertFalse("__name__" in self.serialize(self.portal.doc1))
        self.assertFalse("manage_options" in self.serialize(self.portal.doc1))

    def test_serialize_file(self):
        self.portal.invokeFactory("File", id="file1", title="File 1")
        self.portal.file1.file = NamedFile(
            data="Spam and eggs", contentType="text/plain", filename="test.txt"
        )

        file_url = self.portal.file1.absolute_url()
        download_url = f"{file_url}/@@download/file"
        self.assertEqual(
            {
                "filename": "test.txt",
                "content-type": "text/plain",
                "download": download_url,
                "size": 13,
            },
            self.serialize(self.portal.file1).get("file"),
        )

    def test_serialize_empty_file_returns_none(self):
        self.portal.invokeFactory("File", id="file1", title="File 1")

        self.assertEqual(None, self.serialize(self.portal.file1).get("file"))

    def test_serialize_image(self):
        self.portal.invokeFactory("Image", id="image1", title="Image 1")
        image_file = os.path.join(os.path.dirname(__file__), "image.png")
        with open(image_file, "rb") as f:
            image_data = f.read()
        self.portal.image1.image = NamedBlobImage(
            data=image_data, contentType="image/png", filename="image.png"
        )

        self.maxDiff = 99999

        with patch.object(storage, "uuid4", return_value="uuid_1"):
            obj_url = self.portal.image1.absolute_url()
            scale_url_uuid = "uuid_1"
            download_url = f"{obj_url}/@@images/{scale_url_uuid}.png"
            scales = {
                "listing": {"download": download_url, "width": 16, "height": 4},
                "icon": {"download": download_url, "width": 32, "height": 8},
                "tile": {"download": download_url, "width": 64, "height": 16},
                "thumb": {"download": download_url, "width": 128, "height": 33},
                "mini": {"download": download_url, "width": 200, "height": 52},
                "preview": {"download": download_url, "width": 215, "height": 56},
                "large": {"download": download_url, "width": 215, "height": 56},
            }
            if HAS_PLONE_6:
                # PLIP #3279 amended the image scales
                # https://github.com/plone/Products.CMFPlone/pull/3450
                scales["great"] = {
                    "download": download_url,
                    "height": 56,
                    "width": 215,
                }
                scales["huge"] = {
                    "download": download_url,
                    "height": 56,
                    "width": 215,
                }
                scales["larger"] = {
                    "download": download_url,
                    "height": 56,
                    "width": 215,
                }
                scales["large"] = {
                    "download": download_url,
                    "height": 56,
                    "width": 215,
                }
                scales["teaser"] = {
                    "download": download_url,
                    "height": 56,
                    "width": 215,
                }
            self.assertEqual(
                {
                    "filename": "image.png",
                    "content-type": "image/png",
                    "size": 1185,
                    "download": download_url,
                    "width": 215,
                    "height": 56,
                    "scales": scales,
                },
                self.serialize(self.portal.image1)["image"],
            )

    def test_serialize_empty_image_returns_none(self):
        self.portal.invokeFactory("Image", id="image1", title="Image 1")
        self.assertEqual(None, self.serialize(self.portal.image1)["image"])

    def test_serialize_to_json_collection(self):
        self.portal.invokeFactory("Collection", id="collection1")
        self.portal.collection1.title = "My Collection"
        self.portal.collection1.description = "This is a collection with two documents"
        self.portal.collection1.query = [
            {
                "i": "portal_type",
                "o": "plone.app.querystring.operation.string.is",
                "v": "Document",
            }
        ]
        self.portal.invokeFactory("Document", id="doc2", title="Document 2")
        self.portal.doc1.reindexObject()
        self.portal.doc2.reindexObject()

        self.assertEqual(
            "Collection", self.serialize(self.portal.collection1).get("@type")
        )
        self.assertEqual(
            "Collection", self.serialize(self.portal.collection1).get("@type")
        )
        self.assertEqual(
            [
                {
                    "@id": self.portal.doc1.absolute_url(),
                    "@type": "Document",
                    "description": "",
                    "title": "Document 1",
                    "review_state": "private",
                },
                {
                    "@id": self.portal.doc2.absolute_url(),
                    "@type": "Document",
                    "description": "",
                    "title": "Document 2",
                    "review_state": "private",
                },
            ],
            self.serialize(self.portal.collection1).get("items"),
        )

    def test_serialize_to_json_collection_fullobjects(self):
        self.portal.invokeFactory("Collection", id="collection1")
        self.portal.collection1.title = "My Collection"
        self.portal.collection1.description = "This is a collection with two documents"
        self.portal.collection1.query = [
            {
                "i": "portal_type",
                "o": "plone.app.querystring.operation.string.is",
                "v": "Document",
            }
        ]
        self.portal.invokeFactory("Document", id="doc2", title="Document 2")
        self.portal.doc1.reindexObject()
        self.portal.doc2.reindexObject()

        self.assertEqual(
            "Collection", self.serialize(self.portal.collection1).get("@type")
        )
        self.assertEqual(
            "Collection", self.serialize(self.portal.collection1).get("@type")
        )

        items = self.serialize(self.portal.collection1, fullobjects=True).get("items")
        self.assertIn("UID", items[0])
        self.assertEqual(items[0]["id"], self.portal.doc1.getId())

        self.assertIn("UID", items[1])
        self.assertEqual(items[1]["id"], self.portal.doc2.getId())

    def test_serialize_to_json_collection_include_items(self):
        self.portal.invokeFactory("Collection", id="collection1")
        self.portal.collection1.title = "My Collection"
        self.portal.collection1.description = "This is a collection with two documents"
        self.portal.collection1.query = [
            {
                "i": "portal_type",
                "o": "plone.app.querystring.operation.string.is",
                "v": "Document",
            }
        ]
        self.portal.invokeFactory("Document", id="doc2", title="Document 2")
        self.portal.doc1.reindexObject()
        self.portal.doc2.reindexObject()

        self.assertEqual(
            "Collection", self.serialize(self.portal.collection1).get("@type")
        )
        self.assertEqual(
            "Collection", self.serialize(self.portal.collection1).get("@type")
        )

        self.request.form["include_items"] = False
        without_items = self.serialize(self.portal.collection1)
        self.assertFalse("items" in without_items)
        self.assertFalse("items_total" in without_items)

        self.request.form["include_items"] = True
        serialized = self.serialize(self.portal.collection1)
        items = serialized.get("items")
        self.assertEqual(
            sorted([item["title"] for item in items]),
            [self.portal.doc1.Title(), self.portal.doc2.Title()],
        )
        self.assertEqual(serialized.get("items_total"), 2)

    def test_serialize_returns_site_root_common(self):
        self.assertIn("title", self.serialize(self.portal))
        self.assertIn("description", self.serialize(self.portal))

    @unittest.skipIf(
        HAS_PLONE_6,
        "This test is only intended to run for Plone 6 and DX site root enabled",
    )
    def test_serialize_returns_site_root_opt_in_blocks_not_present(self):
        self.assertEqual(self.serialize(self.portal)["blocks"], {})
        self.assertEqual(self.serialize(self.portal)["blocks_layout"], {})

    @unittest.skipIf(
        HAS_PLONE_6,
        "This test is only intended to run for Plone 6 and DX site root enabled",
    )
    def test_serialize_returns_site_root_opt_in_blocks_present(self):
        blocks = {
            "0358abe2-b4f1-463d-a279-a63ea80daf19": {"@type": "description"},
            "07c273fc-8bfc-4e7d-a327-d513e5a945bb": {"@type": "title"},
        }
        blocks_layout = {
            "items": [
                "07c273fc-8bfc-4e7d-a327-d513e5a945bb",
                "0358abe2-b4f1-463d-a279-a63ea80daf19",
            ]
        }
        self.portal.manage_addProperty("blocks", json.dumps(blocks), "string")
        self.portal.manage_addProperty(
            "blocks_layout", json.dumps(blocks_layout), "string"
        )

        self.assertEqual(self.serialize(self.portal)["blocks"], blocks)
        self.assertEqual(self.serialize(self.portal)["blocks_layout"], blocks_layout)



from plone.app.testing import login
from plone.app.testing import logout
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.dexterity.utils import iterSchemata
from plone.namedfile.file import NamedBlobImage
from plone.namedfile.file import NamedFile
from plone.restapi.interfaces import IFieldDeserializer
from plone.restapi.interfaces import IFieldSerializer
from plone.restapi.testing import PLONE_RESTAPI_BLOCKS_FUNCTIONAL_TESTING
from plone.restapi.testing import PLONE_RESTAPI_BLOCKS_INTEGRATION_TESTING
from plone.uuid.interfaces import IUUID
from Products.CMFCore.utils import getToolByName
from unittest import TestCase
from z3c.form.interfaces import IDataManager
from zope.component import getMultiAdapter

import os
import requests
import transaction


class TestBlocksResolveUIDFunctional(TestCase):

    layer = PLONE_RESTAPI_BLOCKS_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Member"])
        login(self.portal, SITE_OWNER_NAME)
        self.portal.invokeFactory("Folder", id="folder1", title="My Folder")
        self.portal.invokeFactory("Document", id="target", title="Link Target")
        wftool = getToolByName(self.portal, "portal_workflow")
        wftool.doActionFor(self.portal.folder1, "publish")
        transaction.commit()

    def test_create_document_with_link_stores_uuid(self):
        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={
                "@type": "Document",
                "id": "mydocument",
                "title": "My Document",
                "blocks": {
                    "09e39ddf-a945-49f2-b609-ea427ac3430b": {
                        "@type": "text",
                        "text": {
                            "blocks": [
                                {
                                    "data": {},
                                    "depth": 0,
                                    "entityRanges": [
                                        {"key": 0, "length": 4, "offset": 0}
                                    ],
                                    "inlineStyleRanges": [],
                                    "key": "ark35",
                                    "text": "Link",
                                    "type": "unstyled",
                                }
                            ],
                            "entityMap": {
                                "0": {
                                    "data": {"url": f"{self.portal_url}/target"},
                                    "mutability": "MUTABLE",
                                    "type": "LINK",
                                }
                            },
                        },
                    },
                    "21270e22-3a61-4780-b164-d6be56d942f4": {"@type": "title"},
                },
                "blocks_layout": {
                    "items": [
                        "21270e22-3a61-4780-b164-d6be56d942f4",
                        "09e39ddf-a945-49f2-b609-ea427ac3430b",
                    ]
                },
            },
        )
        self.assertEqual(201, response.status_code)
        transaction.begin()

        target_uuid = IUUID(self.portal.target)
        self.assertEqual(
            f"../../resolveuid/{target_uuid}",
            self.portal.folder1.mydocument.blocks.get(
                "09e39ddf-a945-49f2-b609-ea427ac3430b"
            )
            .get("text")
            .get("entityMap")
            .get("0")
            .get("data")
            .get("url"),
        )

    def test_create_document_with_image_block_stores_uuid(self):
        self.portal.invokeFactory("Image", id="image", title="Image")
        image_file = os.path.join(os.path.dirname(__file__), "image.png")
        with open(image_file, "rb") as f:
            image_data = f.read()
        self.portal.image.image = NamedBlobImage(
            data=image_data, contentType="image/png", filename="image.png"
        )
        self.portal.image.image_caption = "This is an image caption."
        transaction.commit()

        target_uuid = IUUID(self.portal.image)

        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={
                "@type": "Document",
                "id": "mydocument",
                "title": "My Document",
                "blocks": {
                    "09e39ddf-a945-49f2-b609-ea427ac3430b": {
                        "@type": "image",
                        "url": f"{self.portal_url}/image",
                    },
                    "21270e22-3a61-4780-b164-d6be56d942f4": {"@type": "title"},
                },
                "blocks_layout": {
                    "items": [
                        "21270e22-3a61-4780-b164-d6be56d942f4",
                        "09e39ddf-a945-49f2-b609-ea427ac3430b",
                    ]
                },
            },
        )
        self.assertEqual(201, response.status_code)
        transaction.begin()

        self.assertEqual(
            f"../../resolveuid/{target_uuid}",
            self.portal.folder1.mydocument.blocks.get(
                "09e39ddf-a945-49f2-b609-ea427ac3430b"
            ).get("url"),
        )

    def test_create_document_with_image_block_and_href_stores_uuid(self):
        self.portal.invokeFactory("Document", id="linked_document", title="Linked Doc")
        self.portal.invokeFactory("Image", id="image", title="Image")
        image_file = os.path.join(os.path.dirname(__file__), "image.png")
        with open(image_file, "rb") as f:
            image_data = f.read()
        self.portal.image.image = NamedBlobImage(
            data=image_data, contentType="image/png", filename="image.png"
        )
        self.portal.image.image_caption = "This is an image caption."
        transaction.commit()

        target_uuid = IUUID(self.portal.image)
        liked_doc_uuid = IUUID(self.portal.linked_document)

        response = requests.post(
            self.portal.folder1.absolute_url(),
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={
                "@type": "Document",
                "id": "mydocument",
                "title": "My Document",
                "blocks": {
                    "09e39ddf-a945-49f2-b609-ea427ac3430b": {
                        "@type": "image",
                        "url": f"{self.portal_url}/image",
                        "href": f"{self.portal_url}/linked_document",
                    },
                    "21270e22-3a61-4780-b164-d6be56d942f4": {"@type": "title"},
                },
                "blocks_layout": {
                    "items": [
                        "21270e22-3a61-4780-b164-d6be56d942f4",
                        "09e39ddf-a945-49f2-b609-ea427ac3430b",
                    ]
                },
            },
        )
        self.assertEqual(201, response.status_code)
        transaction.begin()

        self.assertEqual(
            f"../../resolveuid/{target_uuid}",
            self.portal.folder1.mydocument.blocks.get(
                "09e39ddf-a945-49f2-b609-ea427ac3430b"
            ).get("url"),
        )
        self.assertEqual(
            f"../../resolveuid/{liked_doc_uuid}",
            self.portal.folder1.mydocument.blocks.get(
                "09e39ddf-a945-49f2-b609-ea427ac3430b"
            ).get("href"),
        )


class TestBlocksResolveUID(TestCase):
    layer = PLONE_RESTAPI_BLOCKS_INTEGRATION_TESTING
    maxDiff = None

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        self.doc1 = self.portal[
            self.portal.invokeFactory(
                "Document", id="doc1", title="Document with Blocks"
            )
        ]
        self.doc2 = self.portal[
            self.portal.invokeFactory("Document", id="doc2", title="Target Document")
        ]

        self.doc_primary_field_url = self.portal[
            self.portal.invokeFactory(
                "DXTestDocument",
                id="doc_primary_field_url",
                title="Target Document with primary file field",
                test_primary_namedfile_field=NamedFile(
                    data="Spam and eggs",
                    contentType="text/plain",
                    filename="test.txt",
                ),
            )
        ]

    def serialize(self, fieldname, value):
        for schema in iterSchemata(self.doc1):
            if fieldname in schema:
                field = schema.get(fieldname)
                break
        dm = getMultiAdapter((self.doc1, field), IDataManager)
        dm.set(value)
        serializer = getMultiAdapter((field, self.doc1, self.request), IFieldSerializer)
        return serializer()

    def deserialize(self, fieldname, value):
        for schema in iterSchemata(self.portal.doc1):
            if fieldname in schema:
                field = schema.get(fieldname)
                break
        deserializer = getMultiAdapter(
            (field, self.portal.doc1, self.request), IFieldDeserializer
        )
        return deserializer(value)

    def test_blocks_field_serialization_resolves_uids(self):
        uid = IUUID(self.doc2)
        blocks = {
            "07c273fc-8bfc-4e7d-a327-d513e5a945bb": {"@type": "title"},
            "effbdcdc-253c-41a7-841e-5edb3b56ce32": {
                "@type": "text",
                "text": {
                    "blocks": [
                        {
                            "data": {},
                            "depth": 0,
                            "entityRanges": [{"key": 0, "length": 5, "offset": 0}],
                            "inlineStyleRanges": [],
                            "key": "68rve",
                            "text": "Volto also supports other APIs.",
                            "type": "unstyled",
                        }
                    ],
                    "entityMap": {
                        "0": {
                            "data": {
                                "rel": "nofollow",
                                "url": f"../resolveuid/{uid}",
                            },
                            "mutability": "MUTABLE",
                            "type": "LINK",
                        }
                    },
                },
            },
        }
        value = self.serialize("blocks", blocks)
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            self.doc2.absolute_url(),
        )
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            self.doc2.absolute_url(),
        )

    def test_resolveuid_keeps_suffix(self):
        uid = IUUID(self.doc2)
        blocks = {
            "effbdcdc-253c-41a7-841e-5edb3b56ce32": {
                "@type": "text",
                "text": {
                    "entityMap": {
                        "0": {
                            "data": {
                                "rel": "nofollow",
                                "url": f"../resolveuid/{uid}/view",
                            },
                            "mutability": "MUTABLE",
                            "type": "LINK",
                        }
                    }
                },
            }
        }
        value = self.serialize("blocks", blocks)
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            self.doc2.absolute_url() + "/view",
        )
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            self.doc2.absolute_url() + "/view",
        )

    def test_resolveuid_gets_serialized_for_standard_fields(self):
        uid = IUUID(self.doc2)
        blocks = {"aaa": {"@type": "foo", "url": f"../resolveuid/{uid}/view"}}
        value = self.serialize("blocks", blocks)
        self.assertEqual(value["aaa"]["url"], self.doc2.absolute_url() + "/view")

        blocks = {"aaa": {"@type": "foo", "href": f"../resolveuid/{uid}/view"}}
        value = self.serialize("blocks", blocks)
        self.assertEqual(value["aaa"]["href"], self.doc2.absolute_url() + "/view")

    def test_resolveuid_serialize_take_care_of_primary_fields(self):
        logout()
        uid = IUUID(self.doc_primary_field_url)
        blocks = {"aaa": {"@type": "foo", "url": f"../resolveuid/{uid}"}}
        value = self.serialize("blocks", blocks)
        self.assertEqual(
            value["aaa"]["url"],
            self.doc_primary_field_url.absolute_url()
            + "/@@download/test_primary_namedfile_field",
        )

    def test_keeps_resolveuid_link_if_unknown_uid(self):
        uid = "0000"
        blocks = {
            "effbdcdc-253c-41a7-841e-5edb3b56ce32": {
                "@type": "text",
                "text": {
                    "entityMap": {
                        "0": {
                            "data": {
                                "href": f"../resolveuid/{uid}",
                                "rel": "nofollow",
                                "url": f"../resolveuid/{uid}",
                            },
                            "mutability": "MUTABLE",
                            "type": "LINK",
                        }
                    }
                },
            }
        }
        value = self.serialize("blocks", blocks)
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["href"],
            f"../resolveuid/{uid}",
        )
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            f"../resolveuid/{uid}",
        )

    def test_blocks_field_serialization_doesnt_update_stored_values(self):
        uid = IUUID(self.doc2)
        blocks = {
            "07c273fc-8bfc-4e7d-a327-d513e5a945bb": {"@type": "title"},
            "effbdcdc-253c-41a7-841e-5edb3b56ce32": {
                "@type": "text",
                "text": {
                    "blocks": [
                        {
                            "data": {},
                            "depth": 0,
                            "entityRanges": [{"key": 0, "length": 5, "offset": 0}],
                            "inlineStyleRanges": [],
                            "key": "68rve",
                            "text": "Volto also supports other APIs.",
                            "type": "unstyled",
                        }
                    ],
                    "entityMap": {
                        "0": {
                            "data": {
                                "rel": "nofollow",
                                "url": f"../resolveuid/{uid}",
                            },
                            "mutability": "MUTABLE",
                            "type": "LINK",
                        }
                    },
                },
            },
        }
        value = self.serialize("blocks", blocks)
        self.assertNotEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            blocks["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
        )
        self.assertNotEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            blocks["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
        )

    def test_blocks_field_deserialization_resolves_paths_to_uids(self):
        uid = IUUID(self.doc2)
        blocks = {
            "07c273fc-8bfc-4e7d-a327-d513e5a945bb": {"@type": "title"},
            "effbdcdc-253c-41a7-841e-5edb3b56ce32": {
                "@type": "text",
                "text": {
                    "blocks": [
                        {
                            "data": {},
                            "depth": 0,
                            "entityRanges": [{"key": 0, "length": 5, "offset": 0}],
                            "inlineStyleRanges": [],
                            "key": "68rve",
                            "text": "Volto also supports other APIs.",
                            "type": "unstyled",
                        }
                    ],
                    "entityMap": {
                        "0": {
                            "data": {
                                "url": self.doc2.absolute_url(),
                            },
                            "mutability": "MUTABLE",
                            "type": "LINK",
                        }
                    },
                },
            },
        }
        value = self.deserialize("blocks", blocks)
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            f"../resolveuid/{uid}",
        )
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            f"../resolveuid/{uid}",
        )

    def test_keeps_url_if_unknown_path(self):
        blocks = {
            "effbdcdc-253c-41a7-841e-5edb3b56ce32": {
                "@type": "text",
                "text": {
                    "entityMap": {
                        "0": {
                            "data": {
                                "href": self.portal.absolute_url() + "/foo",
                                "rel": "nofollow",
                                "url": self.portal.absolute_url() + "/foo",
                            },
                            "mutability": "MUTABLE",
                            "type": "LINK",
                        }
                    }
                },
            }
        }
        value = self.deserialize("blocks", blocks)
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["href"],
            self.portal.absolute_url() + "/foo",
        )
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            self.portal.absolute_url() + "/foo",
        )

    def test_path_keeps_suffix(self):
        uid = IUUID(self.doc2)
        blocks = {
            "effbdcdc-253c-41a7-841e-5edb3b56ce32": {
                "@type": "text",
                "text": {
                    "entityMap": {
                        "0": {
                            "data": {
                                "href": self.doc2.absolute_url() + "/view",
                                "rel": "nofollow",
                                "url": self.doc2.absolute_url() + "/view",
                            },
                            "mutability": "MUTABLE",
                            "type": "LINK",
                        }
                    }
                },
            }
        }
        value = self.deserialize("blocks", blocks)
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            f"../resolveuid/{uid}/view",
        )
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            f"../resolveuid/{uid}/view",
        )

    def test_blocks_field_serialization_resolves_uids_with_primary_field_url(self):
        logout()
        uid = IUUID(self.doc_primary_field_url)
        blocks = {
            "07c273fc-8bfc-4e7d-a327-d513e5a945bb": {"@type": "title"},
            "effbdcdc-253c-41a7-841e-5edb3b56ce32": {
                "@type": "text",
                "text": {
                    "blocks": [
                        {
                            "data": {},
                            "depth": 0,
                            "entityRanges": [{"key": 0, "length": 5, "offset": 0}],
                            "inlineStyleRanges": [],
                            "key": "68rve",
                            "text": "Volto also supports other APIs.",
                            "type": "unstyled",
                        }
                    ],
                    "entityMap": {
                        "0": {
                            "data": {
                                "rel": "nofollow",
                                "url": f"../resolveuid/{uid}",
                            },
                            "mutability": "MUTABLE",
                            "type": "LINK",
                        }
                    },
                },
            },
        }
        value = self.serialize("blocks", blocks)
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            self.doc_primary_field_url.absolute_url()
            + "/@@download/test_primary_namedfile_field",
        )
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            self.doc_primary_field_url.absolute_url()
            + "/@@download/test_primary_namedfile_field",
        )

    def test_blocks_field_serialization_resolves_uids_primary_url_with_edit_permission(
        self,
    ):
        uid = IUUID(self.doc_primary_field_url)
        blocks = {
            "07c273fc-8bfc-4e7d-a327-d513e5a945bb": {"@type": "title"},
            "effbdcdc-253c-41a7-841e-5edb3b56ce32": {
                "@type": "text",
                "text": {
                    "blocks": [
                        {
                            "data": {},
                            "depth": 0,
                            "entityRanges": [{"key": 0, "length": 5, "offset": 0}],
                            "inlineStyleRanges": [],
                            "key": "68rve",
                            "text": "Volto also supports other APIs.",
                            "type": "unstyled",
                        }
                    ],
                    "entityMap": {
                        "0": {
                            "data": {
                                "rel": "nofollow",
                                "url": f"../resolveuid/{uid}",
                            },
                            "mutability": "MUTABLE",
                            "type": "LINK",
                        }
                    },
                },
            },
        }
        value = self.serialize("blocks", blocks)
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            self.doc_primary_field_url.absolute_url(),
        )
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            self.doc_primary_field_url.absolute_url(),
        )

    def test_resolveuid_with_primary_field_url_keeps_suffix(self):
        uid = IUUID(self.doc2)
        blocks = {
            "effbdcdc-253c-41a7-841e-5edb3b56ce32": {
                "@type": "text",
                "text": {
                    "entityMap": {
                        "0": {
                            "data": {
                                "rel": "nofollow",
                                "url": f"../resolveuid/{uid}/view",
                            },
                            "mutability": "MUTABLE",
                            "type": "LINK",
                        }
                    }
                },
            }
        }
        value = self.serialize("blocks", blocks)
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            self.doc2.absolute_url() + "/view",
        )
        self.assertEqual(
            value["effbdcdc-253c-41a7-841e-5edb3b56ce32"]["text"]["entityMap"]["0"][
                "data"
            ]["url"],
            self.doc2.absolute_url() + "/view",
        )



from importlib import import_module
from plone.restapi.deserializer import json_body
from plone.restapi.deserializer.dxcontent import DeserializeFromJson
from plone.restapi.interfaces import IBlockFieldDeserializationTransformer
from plone.restapi.interfaces import IDeserializeFromJson
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.component import adapter
from zope.component import subscribers
from zope.interface import implementer
from zope.publisher.interfaces import IRequest

import json


HAS_PLONE_6 = getattr(
    import_module("Products.CMFPlone.factory"), "PLONE60MARKER", False
)


@implementer(IDeserializeFromJson)
@adapter(IPloneSiteRoot, IRequest)
class DeserializeSiteRootFromJson(DeserializeFromJson):
    """JSON deserializer for the Plone site root"""

    def __init__(self, context, request):
        self.context = context
        self.request = request

        if HAS_PLONE_6:
            super().__init__(self.context, self.request)

    def __call__(self, validate_all=False):
        if HAS_PLONE_6:
            # Call the default DX content deserializer
            super().__call__(self)

        data = json_body(self.request)

        if "layout" in data:
            layout = data["layout"]
            self.context.setLayout(layout)

        # Volto Blocks on the Plone Site root faker for Plone 5
        if not HAS_PLONE_6:
            # OrderingMixin (needed for correct ordering for Plone < 6)
            if "ordering" in data and "subset_ids" not in data["ordering"]:
                data["ordering"]["subset_ids"] = self.context.contentIds()
            self.handle_ordering(data)

            if "blocks" in data:
                value = data["blocks"]
                for id, block_value in value.items():
                    block_type = block_value.get("@type", "")
                    handlers = []
                    for h in subscribers(
                        (self.context, self.request),
                        IBlockFieldDeserializationTransformer,
                    ):
                        if h.block_type == block_type or h.block_type is None:
                            handlers.append(h)
                    for handler in sorted(handlers, key=lambda h: h.order):
                        block_value = handler(block_value)
                    value[id] = block_value
                if not getattr(self.context, "blocks", False):
                    self.context.manage_addProperty(
                        "blocks", json.dumps(value), "string"
                    )  # noqa
                else:
                    self.context.manage_changeProperties(
                        blocks=json.dumps(value)
                    )  # noqa

            if "blocks_layout" in data:
                if not getattr(self.context, "blocks_layout", False):
                    self.context.manage_addProperty(
                        "blocks_layout", json.dumps(data["blocks_layout"]), "string"
                    )  # noqa
                else:
                    self.context.manage_changeProperties(
                        blocks_layout=json.dumps(data["blocks_layout"])
                    )  # noqa

            if "title" in data:
                self.context.setTitle(data["title"])

            if "description" in data:
                self.context.manage_changeProperties(
                    description=data["description"]
                )  # noqa

        return self.context



from Acquisition import aq_base
from plone import api
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.restapi.serializer.local_roles import SerializeLocalRolesToJson
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from Products.CMFCore.utils import getToolByName
from zope.component import getGlobalSiteManager

import requests
import transaction
import unittest


def sorted_roles(roles):
    results = []
    for line in roles:
        line = list(line)
        line[1] = sorted(line[1])
        results.append(line)
    return results


class TestFolderCreate(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Member"])
        login(self.portal, SITE_OWNER_NAME)
        self.portal.invokeFactory("Folder", id="folder1", title="My Folder")
        wftool = getToolByName(self.portal, "portal_workflow")
        wftool.doActionFor(self.portal.folder1, "publish")

        self.portal.folder1.invokeFactory("Document", id="doc1", title="My Document")

        transaction.commit()

    def _get_ac_local_roles_block(self, obj):
        return bool(
            getattr(aq_base(self.portal.folder1), "__ac_local_roles_block__", False)
        )

    def test_sharing_search(self):
        """A request to @sharing should support the search parameter."""
        response = requests.get(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )
        non_search_entries = response.json()["entries"]

        response = requests.get(
            self.portal.folder1.absolute_url() + "/@sharing?search=admin",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )
        search_entries = response.json()["entries"]

        # Did we find anything?
        self.assertNotEqual(len(non_search_entries), len(search_entries))

    def test_sharing_search_roundtrip(self):
        """Search for a user and use save roles"""
        # Make sure we don't already have admin
        response = requests.get(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )
        self.assertNotIn("admin", [x["id"] for x in response.json()["entries"]])

        # Now find admin and set roles
        response = requests.get(
            self.portal.folder1.absolute_url() + "/@sharing?search=admin",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )
        roles = [x for x in response.json()["entries"] if x["id"] == "admin"]
        roles = roles[0]["roles"]

        new_roles = {key: not val for key, val in roles.items()}
        payload = {"entries": [{"id": "admin", "roles": new_roles}]}

        response = requests.post(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json=payload,
        )
        self.assertEqual(204, response.status_code)

        # Now we should have admin in @sharing
        response = requests.get(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )
        self.assertIn("admin", [x["id"] for x in response.json()["entries"]])

        # with the same roles as set
        roles = [x for x in response.json()["entries"] if x["id"] == "admin"]
        roles = roles[0]["roles"]
        self.assertEqual(new_roles, roles)

    def test_sharing_titles_are_translated(self):
        response = requests.get(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json", "Accept-Language": "de"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )
        available_roles = response.json()["available_roles"]
        self.assertEqual(
            [
                {"id": "Contributor", "title": "Kann hinzuf\xfcgen"},
                {"id": "Editor", "title": "Kann bearbeiten"},
                {"id": "Reader", "title": "Kann ansehen"},
                {"id": "Reviewer", "title": "Kann ver\xf6ffentlichen"},
            ],
            available_roles,
        )

    def test_sharing_requires_delegate_roles_permission(self):
        """A response for an object without any roles assigned"""
        response = requests.get(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
        )

        self.assertEqual(response.status_code, 403)

    def test_get_local_roles_none_assigned(self):
        """A response for an object without any roles assigned"""
        response = requests.get(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "available_roles": [
                    {"id": "Contributor", "title": "Can add"},
                    {"id": "Editor", "title": "Can edit"},
                    {"id": "Reader", "title": "Can view"},
                    {"id": "Reviewer", "title": "Can review"},
                ],
                "entries": [
                    {
                        "disabled": False,
                        "id": "AuthenticatedUsers",
                        "login": None,
                        "roles": {
                            "Contributor": False,
                            "Editor": False,
                            "Reader": False,
                            "Reviewer": False,
                        },
                        "title": "Logged-in users",
                        "type": "group",
                    }
                ],
                "inherit": True,
            },
        )

    def test_get_local_roles_with_user(self):
        api.user.grant_roles(
            username=TEST_USER_ID, obj=self.portal.folder1, roles=["Reviewer"]
        )
        transaction.commit()

        response = requests.get(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.json(),
            {
                "available_roles": [
                    {"id": "Contributor", "title": "Can add"},
                    {"id": "Editor", "title": "Can edit"},
                    {"id": "Reader", "title": "Can view"},
                    {"id": "Reviewer", "title": "Can review"},
                ],
                "entries": [
                    {
                        "disabled": False,
                        "id": "AuthenticatedUsers",
                        "login": None,
                        "roles": {
                            "Contributor": False,
                            "Editor": False,
                            "Reader": False,
                            "Reviewer": False,
                        },
                        "title": "Logged-in users",
                        "type": "group",
                    },
                    {
                        "disabled": False,
                        "id": "test_user_1_",
                        "roles": {
                            "Contributor": False,
                            "Editor": False,
                            "Reader": False,
                            "Reviewer": True,
                        },
                        "title": "test-user",
                        "type": "user",
                    },
                ],
                "inherit": True,
            },
        )

    def test_set_local_roles_for_user(self):

        pas = getToolByName(self.portal, "acl_users")
        self.assertEqual(
            pas.getLocalRolesForDisplay(self.portal.folder1),
            (("admin", ("Owner",), "user", "admin"),),
        )

        response = requests.post(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={
                "entries": [
                    {
                        "id": TEST_USER_ID,
                        "roles": {
                            "Contributor": False,
                            "Editor": False,
                            "Reader": True,
                            "Reviewer": True,
                        },
                        "type": "user",
                    }
                ]
            },
        )

        transaction.commit()

        self.assertEqual(response.status_code, 204)
        self.assertEqual(
            sorted_roles(pas.getLocalRolesForDisplay(self.portal.folder1)),
            [
                ["admin", ["Owner"], "user", "admin"],
                ["test-user", ["Reader", "Reviewer"], "user", "test_user_1_"],
            ],
        )

    def test_may_only_manage_roles_already_held(self):
        # Grant Editor role to our test user (which gives them the required
        # "plone.DelegateRoles" permission to manage local roles at all)
        api.user.grant_roles(
            username=TEST_USER_ID, obj=self.portal.folder1, roles=["Editor"]
        )
        transaction.commit()

        # Guard assertion - our test user starts with a limited set of roles
        existing_roles = api.user.get_roles(
            username=TEST_USER_ID, obj=self.portal.folder1
        )
        self.assertEqual(
            sorted(["Member", "Authenticated", "Editor"]), sorted(existing_roles)
        )

        # Attempt to gain additional roles not already held
        response = requests.post(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(TEST_USER_NAME, TEST_USER_PASSWORD),
            json={
                "entries": [
                    {
                        "id": TEST_USER_ID,
                        "roles": {
                            "Contributor": True,
                            "Editor": True,
                            "Reader": True,
                            "Publisher": True,
                            "Reviewer": True,
                            "Manager": True,
                        },
                        "type": "user",
                    }
                ]
            },
        )

        transaction.commit()

        self.assertEqual(response.status_code, 204)
        new_roles = api.user.get_roles(username=TEST_USER_ID, obj=self.portal.folder1)

        # New roles should not contain any new roles that the user didn't
        # have permission to delegate.
        self.assertNotIn("Manager", new_roles)
        self.assertNotIn("Publisher", new_roles)
        self.assertNotIn("Reviewer", new_roles)
        self.assertNotIn("Contributor", new_roles)

        # 'Reader' gets added because the permission to delegate it is
        # assigned to 'Editor' by default (see p.a.workflow.permissions)
        self.assertEqual(
            sorted(["Member", "Authenticated", "Editor", "Reader"]), sorted(new_roles)
        )

    def test_unmanaged_existing_roles_are_retained_on_update(self):
        """Make sure that existing roles don't get dropped when a user that
        doesn't manage that roles updates local roles for another user that
        already holds that role.
        """
        # Create another user that holds the Reviewer role, which is not
        # managed by our test user
        api.user.create(
            username="peter",
            email="peter@example.org",
            password="secret",
            roles=("Member",),
        )

        api.user.grant_roles(
            username="peter", obj=self.portal.folder1, roles=["Reviewer"]
        )
        transaction.commit()

        peters_existing_roles = api.user.get_roles(
            username="peter", obj=self.portal.folder1
        )
        self.assertEqual(
            sorted(["Member", "Reviewer", "Authenticated"]),
            sorted(peters_existing_roles),
        )

        # Grant Editor role to our test user (which gives them the required
        # "plone.DelegateRoles" permission to manage local roles at all)
        api.user.grant_roles(
            username=TEST_USER_ID, obj=self.portal.folder1, roles=["Editor"]
        )
        transaction.commit()

        # Guard assertion - our test user doesn't have/manage Reviewer
        existing_roles = api.user.get_roles(
            username=TEST_USER_ID, obj=self.portal.folder1
        )
        self.assertEqual(
            sorted(["Member", "Authenticated", "Editor"]), sorted(existing_roles)
        )

        # Test user now gives Editor to peter. This should not lead to
        # peter losing the Reviewer role.
        response = requests.post(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(TEST_USER_NAME, TEST_USER_PASSWORD),
            json={
                "entries": [
                    {
                        "id": "peter",
                        "roles": {
                            "Contributor": False,
                            "Editor": True,
                            "Reader": True,
                            "Publisher": False,
                            "Reviewer": True,
                            "Manager": False,
                        },
                        "type": "user",
                    }
                ]
            },
        )

        transaction.commit()

        self.assertEqual(response.status_code, 204)
        new_roles = api.user.get_roles(username="peter", obj=self.portal.folder1)

        self.assertIn("Reviewer", new_roles)
        self.assertEqual(
            sorted(["Member", "Authenticated", "Editor", "Reader", "Reviewer"]),
            sorted(new_roles),
        )

    def test_unset_local_roles_for_user(self):
        api.user.grant_roles(
            username=TEST_USER_ID, obj=self.portal.folder1, roles=["Reviewer", "Reader"]
        )
        transaction.commit()

        pas = getToolByName(self.portal, "acl_users")
        self.assertEqual(
            sorted_roles(pas.getLocalRolesForDisplay(self.portal.folder1)),
            [
                ["admin", ["Owner"], "user", "admin"],
                ["test-user", ["Reader", "Reviewer"], "user", "test_user_1_"],
            ],
        )

        response = requests.post(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={
                "entries": [
                    {
                        "id": TEST_USER_ID,
                        "roles": {
                            "Contributor": False,
                            "Editor": False,
                            "Reader": False,
                            "Reviewer": True,
                        },
                        "type": "user",
                    }
                ]
            },
        )

        transaction.commit()

        self.assertEqual(response.status_code, 204)
        self.assertEqual(
            pas.getLocalRolesForDisplay(self.portal.folder1),
            (
                ("admin", ("Owner",), "user", "admin"),
                ("test-user", ("Reviewer",), "user", "test_user_1_"),
            ),
        )

    def test_set_local_roles_on_site_root(self):

        pas = getToolByName(self.portal, "acl_users")
        self.assertEqual(
            pas.getLocalRolesForDisplay(self.portal),
            (("admin", ("Owner",), "user", "admin"),),
        )

        response = requests.post(
            self.portal.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={
                "entries": [
                    {
                        "id": TEST_USER_ID,
                        "roles": {
                            "Contributor": False,
                            "Editor": False,
                            "Reader": True,
                            "Reviewer": True,
                        },
                        "type": "user",
                    }
                ]
            },
        )
        transaction.commit()

        self.assertEqual(response.status_code, 204)
        self.assertEqual(
            sorted_roles(pas.getLocalRolesForDisplay(self.portal)),
            [
                ["admin", ["Owner"], "user", "admin"],
                ["test-user", ["Reader", "Reviewer"], "user", "test_user_1_"],
            ],
        )

    def test_get_local_roles_inherit_roles(self):
        # __ac_local_roles_block__ specifies to block inheritance:
        # https://docs.plone.org/develop/plone/security/local_roles.html
        self.portal.folder1.__ac_local_roles_block__ = True
        transaction.commit()

        response = requests.get(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()["inherit"], False)

    def test_set_local_roles_inherit(self):
        self.assertEqual(self._get_ac_local_roles_block(self.portal.folder1), False)

        # block local roles
        response = requests.post(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"inherit": False},
        )

        transaction.commit()
        self.assertEqual(self._get_ac_local_roles_block(self.portal.folder1), True)
        # unblock local roles
        response = requests.post(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
            json={"inherit": True},
        )
        transaction.commit()

        self.assertEqual(response.status_code, 204)
        self.assertEqual(self._get_ac_local_roles_block(self.portal.folder1), False)

    def test_get_available_roles(self):
        api.user.grant_roles(
            username=TEST_USER_ID, obj=self.portal.folder1, roles=["Reviewer"]
        )
        transaction.commit()

        response = requests.get(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )

        self.assertEqual(response.status_code, 200)
        response = response.json()
        self.assertIn("available_roles", response)
        self.assertIn(
            {"id": "Reader", "title": "Can view"}, response["available_roles"]
        )

    def test_inherited_global(self):
        api.user.grant_roles(username=TEST_USER_ID, roles=["Reviewer"])
        api.user.grant_roles(
            username=TEST_USER_ID, obj=self.portal.folder1, roles=["Editor"]
        )
        transaction.commit()

        response = requests.get(
            self.portal.folder1.doc1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )

        response = response.json()
        # find our entry
        entry = [x for x in response["entries"] if x["id"] == TEST_USER_ID][0]

        self.assertEqual("global", entry["roles"]["Reviewer"])
        self.assertEqual("acquired", entry["roles"]["Editor"])

    def test_inherited_global_via_search(self):
        api.user.create(email="jos@henken.local", username="jos")
        api.user.grant_roles(username="jos", roles=["Reviewer"])
        api.user.grant_roles(username="jos", roles=["Editor"], obj=self.portal.folder1)
        transaction.commit()

        response = requests.get(
            self.portal.folder1.doc1.absolute_url() + "/@sharing?search=jos",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )
        response = response.json()
        # find our entry
        entry = [x for x in response["entries"] if x["id"] == "jos"][0]

        self.assertEqual("global", entry["roles"]["Reviewer"])
        self.assertEqual("acquired", entry["roles"]["Editor"])

    def test_no_serializer_available_returns_501(self):
        # This test unregisters the local_roles adapter. The testrunner can
        # not auto-revert this on test tearDown. Therefore if we ever run
        # into test isolation issues. Start to look here first.
        gsm = getGlobalSiteManager()
        gsm.unregisterAdapter(SerializeLocalRolesToJson, name="local_roles")

        response = requests.get(
            self.portal.folder1.absolute_url() + "/@sharing",
            headers={"Accept": "application/json"},
            auth=(SITE_OWNER_NAME, SITE_OWNER_PASSWORD),
        )

        self.assertEqual(response.status_code, 501)
        response = response.json()
        self.assertIn("error", response)
        self.assertEqual("No serializer available.", response["error"]["message"])

        # we need to re-register the adapter here for following tests
        gsm.registerAdapter(SerializeLocalRolesToJson, name="local_roles")



from plone.restapi.batching import HypermediaBatch
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import ISerializeToJsonSummary
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.interface import Interface

import logging

try:
    from ZTUtils.Lazy import Lazy
except ImportError:
    from Products.ZCatalog.Lazy import Lazy

log = logging.getLogger(__name__)


@implementer(ISerializeToJson)
@adapter(Lazy, Interface)
class LazyCatalogResultSerializer:
    """Serializes a ZCatalog resultset (one of the subclasses of `Lazy`) to
    a Python data structure that can in turn be serialized to JSON.
    """

    def __init__(self, lazy_resultset, request):
        self.lazy_resultset = lazy_resultset
        self.request = request

    def __call__(self, fullobjects=False):
        batch = HypermediaBatch(self.request, self.lazy_resultset)

        results = {}
        results["@id"] = batch.canonical_url
        results["items_total"] = batch.items_total
        links = batch.links
        if links:
            results["batching"] = links

        results["items"] = []
        for brain in batch:
            if fullobjects:
                try:
                    obj = brain.getObject()
                except KeyError:
                    # Guard in case the brain returned refers to an object that doesn't
                    # exists because it failed to uncatalog itself or the catalog has
                    # stale cataloged objects for some reason
                    log.warning(
                        "Brain getObject error: {} doesn't exist anymore".format(
                            brain.getPath()
                        )
                    )
                    continue

                result = getMultiAdapter((obj, self.request), ISerializeToJson)(
                    include_items=False
                )
            else:
                result = getMultiAdapter(
                    (brain, self.request), ISerializeToJsonSummary
                )()

            results["items"].append(result)

        return results



from plone.batching.batch import Batch
from plone.restapi.deserializer import json_body
from urllib.parse import parse_qsl
from urllib.parse import urlencode


DEFAULT_BATCH_SIZE = 25


class HypermediaBatch:
    def __init__(self, request, results):
        self.request = request

        self.b_start = int(json_body(self.request).get("b_start", False)) or int(
            self.request.form.get("b_start", 0)
        )
        self.b_size = int(json_body(self.request).get("b_size", False)) or int(
            self.request.form.get("b_size", DEFAULT_BATCH_SIZE)
        )

        self.batch = Batch(results, self.b_size, self.b_start)

    def __iter__(self):
        """Iterate over items in current batch."""
        return iter(self.batch)

    @property
    def items_total(self):
        """Return the number of total items in underlying sequence."""
        return self.batch.sequence_length

    @property
    def canonical_url(self):
        """Return the canonical URL to the batched collection-like resource,
        preserving query string params, but stripping all batching related
        params from it.
        """
        url = self.request["ACTUAL_URL"]
        qs_params = parse_qsl(self.request["QUERY_STRING"])

        # Remove any batching / sorting related parameters.
        # Also take care to preserve list-like query string params.
        for key, value in qs_params[:]:
            if key in ("b_size", "b_start", "sort_on", "sort_order", "sort_limit"):
                qs_params.remove((key, value))

        qs = urlencode(qs_params)

        if qs_params:
            url = "?".join((url, qs))
        return url

    @property
    def current_batch_url(self):
        url = self.request["ACTUAL_URL"]
        qs = self.request["QUERY_STRING"]
        if qs:
            url = "?".join((url, qs))
        return url

    @property
    def links(self):
        """Get a dictionary with batching links."""
        # Don't provide batching links if resultset isn't batched
        if self.items_total <= self.b_size:
            return

        links = {}

        first = self._batch_for_page(1)
        last = self._batch_for_page(self.batch.lastpage)
        next = self.batch.next
        prev = self.batch.previous

        links["@id"] = self.current_batch_url
        links["first"] = self._url_for_batch(first)
        links["last"] = self._url_for_batch(last)

        if next:
            links["next"] = self._url_for_batch(next)

        if prev:
            links["prev"] = self._url_for_batch(prev)

        return links

    def _batch_for_page(self, pagenumber):
        """Return a new Batch object for the given pagenumber."""
        new_batch = Batch.fromPagenumber(
            self.batch._sequence, pagesize=self.b_size, pagenumber=pagenumber
        )
        return new_batch

    def _url_for_batch(self, batch):
        """Return URL that points to the given batch."""
        # Calculate the start for the new batch page.
        # Make sure we account for plone.batching's one-based indexing and
        # that the start never drops below zero
        new_start = max(0, batch.start - 1)
        url = self._url_with_params(params={"b_start": new_start})
        return url

    def _url_with_params(self, params):
        """Build an URL based on the actual URL of the current request URL
        and add or update some query string parameters in it.
        """
        url = self.request["ACTUAL_URL"]
        qs_params = parse_qsl(self.request["QUERY_STRING"], keep_blank_values=1)

        # Take care to preserve list-like query string arguments (same QS
        # param repeated multiple times). In other words, don't turn the
        # result of parse_qsl into a dict!

        # Drop params to be updated, then prepend new params in order
        qs_params = [x for x in qs_params if x[0] not in list(params)]
        qs_params = sorted(params.items()) + qs_params

        qs = urlencode(qs_params)

        if qs_params:
            url = "?".join((url, qs))
        return url



from AccessControl.interfaces import IRoleManager
from Acquisition import aq_base
from operator import itemgetter
from plone.app.workflow.interfaces import ISharingPageRole
from plone.restapi.interfaces import ISerializeToJson
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.component import queryUtility
from zope.i18n import translate
from zope.interface import implementer
from zope.interface import Interface


@adapter(IRoleManager, Interface)
@implementer(ISerializeToJson)
class SerializeLocalRolesToJson:
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def _get_title(self, role_id):
        util = queryUtility(ISharingPageRole, name=role_id)
        if not util:
            return role_id
        return util.title

    def __call__(self, search=None):
        self.request.form["search_term"] = search
        sharing_view = getMultiAdapter((self.context, self.request), name="sharing")
        local_roles = sharing_view.role_settings()

        available_roles = []
        for role in sorted(sharing_view.roles(), key=itemgetter("id")):
            util = queryUtility(ISharingPageRole, name=role["id"])
            title = util.title
            available_roles.append(
                {"id": role["id"], "title": translate(title, context=self.request)}
            )

        blocked_roles = getattr(
            aq_base(self.context), "__ac_local_roles_block__", False
        )

        return {
            "inherit": not blocked_roles,
            "entries": local_roles,
            "available_roles": available_roles,
        }



from datetime import date
from datetime import datetime
from datetime import time
from datetime import timedelta
from decimal import Decimal
from importlib import import_module
from plone.app.textfield.value import RichTextValue
from plone.dexterity.utils import iterSchemata
from plone.namedfile.file import NamedBlobFile
from plone.namedfile.file import NamedBlobImage
from plone.namedfile.file import NamedFile
from plone.namedfile.file import NamedImage
from plone.restapi.interfaces import IFieldSerializer
from plone.restapi.serializer.dxfields import DefaultFieldSerializer
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from plone.scale import storage
from plone.uuid.interfaces import IUUID
from unittest import TestCase
from unittest.mock import patch
from z3c.form.interfaces import IDataManager
from zope.component import getMultiAdapter
from zope.interface.verify import verifyClass

import os


HAS_PLONE_6 = getattr(
    import_module("Products.CMFPlone.factory"), "PLONE60MARKER", False
)


class TestDexterityFieldSerializing(TestCase):
    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING
    maxDiff = None

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        self.doc1 = self.portal[
            self.portal.invokeFactory(
                "DXTestDocument", id="doc1", title="Test Document"
            )
        ]

    def serialize(self, fieldname, value):
        for schema in iterSchemata(self.doc1):
            if fieldname in schema:
                field = schema.get(fieldname)
                break
        dm = getMultiAdapter((self.doc1, field), IDataManager)
        dm.set(value)
        serializer = getMultiAdapter((field, self.doc1, self.request), IFieldSerializer)
        return serializer()

    def test_ascii_field_serialization_returns_unicode(self):
        value = self.serialize("test_ascii_field", "foo")
        self.assertTrue(isinstance(value, str), "Not an <unicode>")
        self.assertEqual("foo", value)

    def test_asciiline_field_serialization_returns_unicode(self):
        value = self.serialize("test_asciiline_field", "foo")
        self.assertTrue(isinstance(value, str), "Not an <unicode>")
        self.assertEqual("foo", value)

    def test_bool_field_serialization_returns_true(self):
        value = self.serialize("test_bool_field", True)
        self.assertTrue(isinstance(value, bool), "Not a <bool>")
        self.assertEqual(True, value)

    def test_bool_field_serialization_returns_false(self):
        value = self.serialize("test_bool_field", False)
        self.assertTrue(isinstance(value, bool), "Not a <bool>")
        self.assertEqual(False, value)

    def test_bytes_field_serialization_returns_unicode(self):
        value = self.serialize("test_bytes_field", b"\xc3\xa4\xc3\xb6\xc3\xbc")
        self.assertTrue(isinstance(value, str), "Not an <unicode>")
        self.assertEqual("\xe4\xf6\xfc", value)

    def test_bytesline_field_serialization_returns_unicode(self):
        value = self.serialize("test_bytesline_field", b"\xc3\xa4\xc3\xb6\xc3\xbc")
        self.assertTrue(isinstance(value, str), "Not an <unicode>")
        self.assertEqual("\xe4\xf6\xfc", value)

    def test_choice_field_serialization_returns_vocabulary_term(self):
        value = self.serialize("test_choice_field", "foo")
        self.assertTrue(isinstance(value, dict))
        self.assertEqual({"token": "foo", "title": None}, value)

    def test_choice_field_with_vocabulary_serialization_returns_vocabulary_term(
        self,
    ):  # noqa
        value = self.serialize("test_choice_field_with_vocabulary", "value1")
        self.assertTrue(isinstance(value, dict))
        self.assertEqual({"token": "token1", "title": "title1"}, value)

    def test_date_field_serialization_returns_unicode(self):
        value = self.serialize("test_date_field", date(2015, 7, 15))
        self.assertTrue(isinstance(value, str), "Not an <unicode>")
        self.assertEqual("2015-07-15", value)

    def test_datetime_field_serialization_returns_unicode(self):
        value = self.serialize("test_datetime_field", datetime(2015, 6, 20, 13, 22, 4))
        self.assertTrue(isinstance(value, str), "Not an <unicode>")
        self.assertEqual("2015-06-20T13:22:04", value)

    def test_decimal_field_serialization_returns_unicode(self):
        value = self.serialize("test_decimal_field", Decimal("1.1"))
        self.assertTrue(isinstance(value, str), "Not an <unicode>")
        self.assertEqual("1.1", value)

    def test_dict_field_serialization_returns_dict(self):
        value = self.serialize(
            "test_dict_field", {"foo": "bar", "spam": "eggs", "1": 1}
        )
        self.assertTrue(isinstance(value, dict), "Not a <dict>")
        self.assertEqual({"foo": "bar", "spam": "eggs", "1": 1}, value)

    def test_float_field_serialization_returns_float(self):
        value = self.serialize("test_float_field", 1.5)
        self.assertTrue(isinstance(value, float), "Not a <float>")
        self.assertEqual(1.5, value)

    def test_frozenset_field_serialization_returns_list(self):
        value = self.serialize("test_frozenset_field", frozenset([1, 2, 3]))
        self.assertTrue(isinstance(value, list), "Not a <list>")
        self.assertEqual([1, 2, 3], sorted(value))

    def test_int_field_serialization_returns_int(self):
        value = self.serialize("test_int_field", 500)
        self.assertTrue(isinstance(value, int), "Not an <int>")
        self.assertEqual(500, value)

    def test_list_field_serialization_returns_list(self):
        value = self.serialize("test_list_field", [1, "two", 3])
        self.assertTrue(isinstance(value, list), "Not a <list>")
        self.assertEqual([1, "two", 3], value)

    def test_list_field_with_vocabulary_choice_serialization_returns_terms(self):
        value = self.serialize(
            "test_list_field_with_choice_with_vocabulary", ["value1", "value3"]
        )
        self.assertTrue(isinstance(value, list), "Not a <list>")
        self.assertEqual(
            [
                {"token": "token1", "title": "title1"},
                {"token": "token3", "title": "title3"},
            ],
            value,
        )

    def test_list_field_with_vocabulary_choice_serialization_no_valid_term(self):
        value = self.serialize(
            "test_list_field_with_choice_with_vocabulary", ["value3", "value4"]
        )
        self.assertTrue(isinstance(value, list), "Not a <list>")
        self.assertEqual(
            [{"token": "token3", "title": "title3"}],
            value,
        )

    def test_set_field_serialization_returns_list(self):
        value = self.serialize("test_set_field", {"a", "b", "c"})
        self.assertTrue(isinstance(value, list), "Not a <list>")
        self.assertEqual(["a", "b", "c"], sorted(value))

    def test_set_field_with_vocabulary_choice_serialization_returns_terms(self):
        value = self.serialize(
            "test_set_field_with_choice_with_vocabulary", {"value1", "value3"}
        )
        self.assertTrue(isinstance(value, list), "Not a <list>")
        self.assertEqual(
            [
                {"token": "token1", "title": "title1"},
                {"token": "token3", "title": "title3"},
            ],
            sorted(value, key=lambda x: x["token"]),
        )

    def test_text_field_serialization_returns_unicode(self):
        value = self.serialize("test_text_field", "Käfer")
        self.assertTrue(isinstance(value, str), "Not an <unicode>")
        self.assertEqual("Käfer", value)

    def test_textline_field_serialization_returns_unicode(self):
        value = self.serialize("test_textline_field", "Käfer")
        self.assertTrue(isinstance(value, str), "Not an <unicode>")
        self.assertEqual("Käfer", value)

    def test_time_field_serialization_returns_unicode(self):
        value = self.serialize("test_time_field", time(14, 15, 33))
        self.assertTrue(isinstance(value, str), "Not an <unicode>")
        self.assertEqual("14:15:33", value)

    def test_timedelta_field_serialization_returns_float(self):
        value = self.serialize("test_timedelta_field", timedelta(0.01))
        self.assertTrue(isinstance(value, float), "Not a <float>")
        self.assertEqual(864.0, value)

    def test_richtext_field_serialization_returns_dict(self):
        value = self.serialize(
            "test_richtext_field",
            RichTextValue(
                raw="<p>Some Text</p>",
                mimeType="text/html",
                outputMimeType="text/html",
            ),
        )
        self.assertTrue(isinstance(value, dict), "Not a <dict>")
        self.assertEqual(
            {
                "content-type": "text/html",
                "data": "<p>Some Text</p>",
                "encoding": "utf-8",
            },
            value,
        )

    def test_namedfile_field_serialization_returns_dict(self):
        value = self.serialize(
            "test_namedfile_field",
            NamedFile(
                data="Spam and eggs", contentType="text/plain", filename="test.txt"
            ),
        )
        self.assertTrue(isinstance(value, dict), "Not a <dict>")
        download_url = "/".join(
            [self.doc1.absolute_url(), "@@download/test_namedfile_field"]
        )
        self.assertEqual(
            {
                "filename": "test.txt",
                "content-type": "text/plain",
                "size": 13,
                "download": download_url,
            },
            value,
        )

    def test_namedblobfile_field_serialization_returns_dict(self):
        value = self.serialize(
            "test_namedblobfile_field",
            NamedBlobFile(
                data="Spam and eggs", contentType="text/plain", filename="test.txt"
            ),
        )
        self.assertTrue(isinstance(value, dict), "Not a <dict>")

        download_url = "/".join(
            [self.doc1.absolute_url(), "@@download/test_namedblobfile_field"]
        )
        self.assertEqual(
            {
                "filename": "test.txt",
                "content-type": "text/plain",
                "size": 13,
                "download": download_url,
            },
            value,
        )

    def test_relationchoice_field_serialization_returns_summary_dict(self):
        doc2 = self.portal[
            self.portal.invokeFactory(
                "DXTestDocument",
                id="doc2",
                title="Referenceable Document",
                description="Description 2",
            )
        ]
        value = self.serialize("test_relationchoice_field", doc2)
        self.assertEqual(
            {
                "@id": "http://nohost/plone/doc2",
                "@type": "DXTestDocument",
                "title": "Referenceable Document",
                "description": "Description 2",
                "review_state": "private",
            },
            value,
        )

    def test_relationlist_field_serialization_returns_list(self):
        doc2 = self.portal[
            self.portal.invokeFactory(
                "DXTestDocument",
                id="doc2",
                title="Referenceable Document",
                description="Description 2",
            )
        ]
        doc3 = self.portal[
            self.portal.invokeFactory(
                "DXTestDocument",
                id="doc3",
                title="Referenceable Document",
                description="Description 3",
            )
        ]
        value = self.serialize("test_relationlist_field", [doc2, doc3])
        self.assertTrue(isinstance(value, list), "Not a <list>")
        self.assertEqual(
            [
                {
                    "@id": "http://nohost/plone/doc2",
                    "@type": "DXTestDocument",
                    "title": "Referenceable Document",
                    "description": "Description 2",
                    "review_state": "private",
                },
                {
                    "@id": "http://nohost/plone/doc3",
                    "@type": "DXTestDocument",
                    "title": "Referenceable Document",
                    "description": "Description 3",
                    "review_state": "private",
                },
            ],
            value,
        )

    def test_remoteurl_field_in_links_get_converted(self):
        link = self.portal[
            self.portal.invokeFactory(
                "Link",
                id="link",
                title="Test Link",
            )
        ]
        field = None
        for schema in iterSchemata(link):
            if "remoteUrl" in schema:
                field = schema.get("remoteUrl")
                break
        dm = getMultiAdapter((link, field), IDataManager)
        serializer = getMultiAdapter((field, link, self.request), IFieldSerializer)

        dm.set("http://www.plone.com")
        self.assertEqual(serializer(), "http://www.plone.com")

        doc_uuid = IUUID(self.portal.doc1)

        dm.set(f"../resolveuid/{doc_uuid}")
        self.assertEqual(serializer(), self.portal.doc1.absolute_url())

        # Support for variable interpolation is still present
        dm.set("${portal_url}/doc1")
        self.assertEqual(serializer(), self.portal.doc1.absolute_url())

        dm.set("/doc1")
        self.assertEqual(serializer(), "/doc1")

        dm.set("/doc2")
        self.assertEqual(serializer(), "/doc2")


class TestDexterityImageFieldSerializingOriginalAndPNGScales(TestCase):
    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING
    maxDiff = None

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        self.doc1 = self.portal[
            self.portal.invokeFactory(
                "DXTestDocument", id="doc1", title="Test Document"
            )
        ]

    def serialize(self, fieldname, value):
        for schema in iterSchemata(self.doc1):
            if fieldname in schema:
                field = schema.get(fieldname)
                break
        dm = getMultiAdapter((self.doc1, field), IDataManager)
        dm.set(value)
        serializer = getMultiAdapter((field, self.doc1, self.request), IFieldSerializer)
        return serializer()

    def test_namedimage_field_serialization_returns_dict_with_original_scale(self):
        """In Plone >= 5.2 the image returned when requesting an image
        scale with the same width and height of the original image is
        the actual original image in its original format"""
        image_file = os.path.join(os.path.dirname(__file__), "1024x768.gif")
        with open(image_file, "rb") as f:
            image_data = f.read()
        fn = "test_namedimage_field"
        with patch.object(storage, "uuid4", return_value="uuid_1"):
            value = self.serialize(
                fn,
                NamedImage(
                    data=image_data, contentType="image/gif", filename="1024x768.gif"
                ),
            )
            self.assertTrue(isinstance(value, dict), "Not a <dict>")

            scale_url_uuid = "uuid_1"
            obj_url = self.doc1.absolute_url()

            # Original image is still a "scale"
            # scaled images are converted to PNG in Plone = 5.2
            original_download_url = "{}/@@images/{}.{}".format(
                obj_url, scale_url_uuid, "gif"
            )

            scale_download_url = "{}/@@images/{}.{}".format(
                obj_url, scale_url_uuid, "png"
            )
            scales = {
                "listing": {
                    "download": scale_download_url,
                    "width": 16,
                    "height": 12,
                },
                "icon": {"download": scale_download_url, "width": 32, "height": 24},
                "tile": {"download": scale_download_url, "width": 64, "height": 48},
                "thumb": {
                    "download": scale_download_url,
                    "width": 128,
                    "height": 96,
                },
                "mini": {
                    "download": scale_download_url,
                    "width": 200,
                    "height": 150,
                },
                "preview": {
                    "download": scale_download_url,
                    "width": 400,
                    "height": 300,
                },
                "large": {
                    "download": scale_download_url,
                    "width": 768,
                    "height": 576,
                },
            }
            if HAS_PLONE_6:
                # PLIP #3279 amended the image scales
                # https://github.com/plone/Products.CMFPlone/pull/3450
                scales["great"] = {
                    "download": scale_download_url,
                    "height": 768,
                    "width": 1024,
                }
                scales["huge"] = {
                    "download": scale_download_url,
                    "height": 768,
                    "width": 1024,
                }
                scales["larger"] = {
                    "download": scale_download_url,
                    "height": 750,
                    "width": 1000,
                }
                scales["large"] = {
                    "download": scale_download_url,
                    "height": 600,
                    "width": 800,
                }
                scales["teaser"] = {
                    "download": scale_download_url,
                    "height": 450,
                    "width": 600,
                }
            self.assertEqual(
                {
                    "filename": "1024x768.gif",
                    "content-type": "image/gif",
                    "size": 1514,
                    "download": original_download_url,
                    "width": 1024,
                    "height": 768,
                    "scales": scales,
                },
                value,
            )

    def test_namedimage_field_serialization_doesnt_choke_on_corrupt_image(self):
        """In Plone >= 5.2 the original image file is not a "scale", so its url
        is returned as is and we need to check it, but the scales should be empty"""
        image_data = b"INVALID IMAGE DATA"
        fn = "test_namedimage_field"
        with patch.object(storage, "uuid4", return_value="uuid_1"):
            value = self.serialize(
                fn,
                NamedImage(
                    data=image_data, contentType="image/gif", filename="1024x768.gif"
                ),
            )

        obj_url = self.doc1.absolute_url()
        scale_url_uuid = "uuid_1"
        self.assertEqual(
            {
                "content-type": "image/gif",
                "download": "{}/@@images/{}.{}".format(obj_url, scale_url_uuid, "gif"),
                "filename": "1024x768.gif",
                "height": -1,
                "scales": {},
                "size": 18,
                "width": -1,
            },
            value,
        )

    def test_namedblobimage_field_serialization_returns_dict_with_original_scale(self):
        """In Plone >= 5.2 the image returned when requesting an image
        scale with the same width and height of the original image is
        the actual original image in its original format"""
        image_file = os.path.join(os.path.dirname(__file__), "1024x768.gif")
        with open(image_file, "rb") as f:
            image_data = f.read()
        fn = "test_namedblobimage_field"
        with patch.object(storage, "uuid4", return_value="uuid_1"):
            value = self.serialize(
                fn,
                NamedBlobImage(
                    data=image_data, contentType="image/gif", filename="1024x768.gif"
                ),
            )
            self.assertTrue(isinstance(value, dict), "Not a <dict>")

            scale_url_uuid = "uuid_1"
            obj_url = self.doc1.absolute_url()

            # Original image is still a "scale"
            # scaled images are converted to PNG in Plone = 5.2
            original_download_url = "{}/@@images/{}.{}".format(
                obj_url, scale_url_uuid, "gif"
            )

            scale_download_url = "{}/@@images/{}.{}".format(
                obj_url, scale_url_uuid, "png"
            )
            scales = {
                "listing": {
                    "download": scale_download_url,
                    "width": 16,
                    "height": 12,
                },
                "icon": {"download": scale_download_url, "width": 32, "height": 24},
                "tile": {"download": scale_download_url, "width": 64, "height": 48},
                "thumb": {
                    "download": scale_download_url,
                    "width": 128,
                    "height": 96,
                },
                "mini": {
                    "download": scale_download_url,
                    "width": 200,
                    "height": 150,
                },
                "preview": {
                    "download": scale_download_url,
                    "width": 400,
                    "height": 300,
                },
                "large": {
                    "download": scale_download_url,
                    "width": 768,
                    "height": 576,
                },
            }
            if HAS_PLONE_6:
                # PLIP #3279 amended the image scales
                # https://github.com/plone/Products.CMFPlone/pull/3450
                scales["great"] = {
                    "download": scale_download_url,
                    "height": 768,
                    "width": 1024,
                }
                scales["huge"] = {
                    "download": scale_download_url,
                    "height": 768,
                    "width": 1024,
                }
                scales["larger"] = {
                    "download": scale_download_url,
                    "height": 750,
                    "width": 1000,
                }
                scales["large"] = {
                    "download": scale_download_url,
                    "height": 600,
                    "width": 800,
                }
                scales["teaser"] = {
                    "download": scale_download_url,
                    "height": 450,
                    "width": 600,
                }
            self.assertEqual(
                {
                    "filename": "1024x768.gif",
                    "content-type": "image/gif",
                    "size": 1514,
                    "download": original_download_url,
                    "width": 1024,
                    "height": 768,
                    "scales": scales,
                },
                value,
            )

    def test_namedblobimage_field_serialization_doesnt_choke_on_corrupt_image(self):
        """In Plone >= 5.2 the original image file is not a "scale", so its url
        is returned as is and we need to check it, but the scales should be empty"""
        image_data = b"INVALID IMAGE DATA"
        fn = "test_namedblobimage_field"
        with patch.object(storage, "uuid4", return_value="uuid_1"):
            value = self.serialize(
                fn,
                NamedBlobImage(
                    data=image_data, contentType="image/gif", filename="1024x768.gif"
                ),
            )

        obj_url = self.doc1.absolute_url()
        scale_url_uuid = "uuid_1"
        self.assertEqual(
            {
                "content-type": "image/gif",
                "download": "{}/@@images/{}.{}".format(obj_url, scale_url_uuid, "gif"),
                "filename": "1024x768.gif",
                "height": -1,
                "scales": {},
                "size": 18,
                "width": -1,
            },
            value,
        )


class TestDexterityFieldSerializers(TestCase):
    def default_field_serializer(self):
        verifyClass(IFieldSerializer, DefaultFieldSerializer)



from plone.dexterity.interfaces import IDexterityContent
from plone.restapi.interfaces import IExpandableElement
from plone.restapi.interfaces import IPloneRestapiLayer
from plone.restapi.services import Service
from plone.restapi.types.utils import get_info_for_type
from plone.restapi.types.utils import get_info_for_field
from plone.restapi.types.utils import get_info_for_fieldset
from Products.CMFCore.interfaces import IFolderish
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces import IConstrainTypes
from zExceptions import Unauthorized
from zope.component import adapter
from zope.component import getMultiAdapter, queryMultiAdapter
from zope.component import getUtility
from zope.i18n import translate
from zope.interface import implementer
from zope.interface import Interface
from zope.interface import noLongerProvides
from zope.publisher.interfaces import IPublishTraverse
from zope.schema.interfaces import IVocabularyFactory


def check_security(context):
    # Only expose type information to authenticated users
    portal_membership = getToolByName(context, "portal_membership")
    if portal_membership.isAnonymousUser():
        raise Unauthorized


@implementer(IExpandableElement)
@adapter(IDexterityContent, Interface)
class TypesInfo:
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self, expand=False):
        result = {"types": {"@id": f"{self.context.absolute_url()}/@types"}}
        if not expand:
            return result

        check_security(self.context)

        vocab_factory = getUtility(
            IVocabularyFactory, name="plone.app.vocabularies.ReallyUserFriendlyTypes"
        )

        portal_types = getToolByName(self.context, "portal_types")
        # allowedContentTypes already checks for permissions
        constrains = IConstrainTypes(self.context, None)
        if constrains:
            allowed_types = constrains.getLocallyAllowedTypes()
            immediately_types = constrains.getImmediatelyAddableTypes()
        else:
            allowed_types = [x.getId() for x in self.context.allowedContentTypes()]
            immediately_types = allowed_types

        portal = getMultiAdapter(
            (self.context, self.request), name="plone_portal_state"
        ).portal()
        portal_url = portal.absolute_url()

        # only addables if the content type is folderish
        can_add = IFolderish.providedBy(self.context)

        # Filter out any type that doesn't have lookupSchema. We are depended
        # on that in lower level code.
        ftis = [portal_types[x.value] for x in vocab_factory(self.context)]
        ftis = [fti for fti in ftis if getattr(fti, "lookupSchema", None)]

        result["types"] = [
            {
                "@id": f"{portal_url}/@types/{fti.getId()}",
                "title": translate(fti.Title(), context=self.request),
                "addable": fti.getId() in allowed_types if can_add else False,
                "immediately_addable": fti.getId() in immediately_types
                if can_add
                else False,
            }
            for fti in ftis
        ]

        return result


@implementer(IPublishTraverse)
class TypesGet(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Treat any path segments after /@types as parameters
        self.params.append(name)
        return self

    @property
    def _get_record_name(self):
        if len(self.params) != 1:
            raise Exception(
                "Must supply exactly one parameter (dotted name of"
                "the record to be retrieved)"
            )

        return self.params[0]

    def reply(self):
        if not self.params:
            # List type info, including addable_types
            info = TypesInfo(self.context, self.request)
            return info(expand=True)["types"]

        if len(self.params) == 1:
            return self.reply_for_type()

        if len(self.params) == 2:
            return self.reply_for_field()

    def reply_for_type(self):
        check_security(self.context)
        portal_type = self.params.pop()

        # Make sure we get the right dexterity-types adapter
        if IPloneRestapiLayer.providedBy(self.request):
            noLongerProvides(self.request, IPloneRestapiLayer)

        try:
            dtool = queryMultiAdapter(
                (self.context, self.request), name="dexterity-types"
            )
            dtype = dtool.publishTraverse(self.request, portal_type)
        except Exception:
            dtype = self.context

        try:
            schema = get_info_for_type(dtype, self.request, portal_type)
        except KeyError:
            self.content_type = "application/json"
            self.request.response.setStatus(404)
            return {
                "type": "NotFound",
                "message": "Type '%s' could not be found." % portal_type,
            }

        self.content_type = "application/json+schema"
        return schema

    def reply_for_field(self):
        check_security(self.context)
        name = self.params[0]
        field_name = self.params[1]

        # Make sure we get the right dexterity-types adapter
        if IPloneRestapiLayer.providedBy(self.request):
            noLongerProvides(self.request, IPloneRestapiLayer)

        context = queryMultiAdapter(
            (self.context, self.request), name="dexterity-types"
        )
        context = context.publishTraverse(self.request, name)

        try:
            # Get field
            return get_info_for_field(context, self.request, field_name)
        except (KeyError, AttributeError):
            # Get fieldset
            return self.reply_for_fieldset()

    def reply_for_fieldset(self):
        name = self.params[0]
        field_name = self.params[1]

        # Make sure we get the right dexterity-types adapter
        if IPloneRestapiLayer.providedBy(self.request):
            noLongerProvides(self.request, IPloneRestapiLayer)

        context = queryMultiAdapter(
            (self.context, self.request), name="dexterity-types"
        )
        context = context.publishTraverse(self.request, name)

        try:
            return get_info_for_fieldset(context, self.request, field_name)
        except KeyError:
            self.content_type = "application/json"
            self.request.response.setStatus(404)
            return {
                "type": "NotFound",
                "message": "Field(set) '%s' could not be found." % field_name,
            }



from AccessControl import getSecurityManager
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.utils import normalizeString
from zExceptions import BadRequest
from zope.component import queryMultiAdapter
from zope.component.hooks import getSite
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse


DEFAULT_SEARCH_RESULTS_LIMIT = 25


@implementer(IPublishTraverse)
class UsersGet(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []
        portal = getSite()
        self.portal_membership = getToolByName(portal, "portal_membership")
        self.acl_users = getToolByName(portal, "acl_users")
        self.query = self.request.form.copy()

    def publishTraverse(self, request, name):
        # Consume any path segments after /@users as parameters
        self.params.append(name)
        return self

    @property
    def _get_user_id(self):
        if len(self.params) != 1:
            raise Exception("Must supply exactly one parameter (user id)")
        return self.params[0]

    def _get_user(self, user_id):
        return self.portal_membership.getMemberById(user_id)

    @staticmethod
    def _sort_users(users):
        users.sort(
            key=lambda x: x is not None
            and normalizeString(x.getProperty("fullname", ""))
        )
        return users

    def _get_users(self):
        results = {user["userid"] for user in self.acl_users.searchUsers()}
        users = [self.portal_membership.getMemberById(userid) for userid in results]
        return self._sort_users(users)

    def _get_filtered_users(self, query, limit):
        results = self.acl_users.searchUsers(id=query, max_results=limit)
        users = [
            self.portal_membership.getMemberById(user["userid"]) for user in results
        ]
        return self._sort_users(users)

    def has_permission_to_query(self):
        sm = getSecurityManager()
        return sm.checkPermission("Manage portal", self.context)

    def has_permission_to_enumerate(self):
        sm = getSecurityManager()
        return sm.checkPermission("Manage portal", self.context)

    def has_permission_to_access_user_info(self):
        sm = getSecurityManager()
        return sm.checkPermission(
            "plone.restapi: Access Plone user information", self.context
        )

    def reply(self):
        if len(self.query) > 0 and len(self.params) == 0:
            query = self.query.get("query", "")
            limit = self.query.get("limit", DEFAULT_SEARCH_RESULTS_LIMIT)
            if query:
                # Someone is searching users, check if they are authorized
                if self.has_permission_to_query():
                    users = self._get_filtered_users(query, limit)
                    result = []
                    for user in users:
                        serializer = queryMultiAdapter(
                            (user, self.request), ISerializeToJson
                        )
                        result.append(serializer())
                    return result
                else:
                    self.request.response.setStatus(401)
                    return
            else:
                raise BadRequest("Query string supplied is not valid")

        if len(self.params) == 0:
            # Someone is asking for all users, check if they are authorized
            if self.has_permission_to_enumerate():
                result = []
                for user in self._get_users():
                    serializer = queryMultiAdapter(
                        (user, self.request), ISerializeToJson
                    )
                    result.append(serializer())
                return result
            else:
                self.request.response.setStatus(401)
                return

        # Some is asking one user, check if the logged in user is asking
        # their own information or if they are a Manager
        current_user_id = self.portal_membership.getAuthenticatedMember().getId()

        if self.has_permission_to_access_user_info() or (
            current_user_id and current_user_id == self._get_user_id
        ):

            # we retrieve the user on the user id not the username
            user = self._get_user(self._get_user_id)
            if not user:
                self.request.response.setStatus(404)
                return
            serializer = queryMultiAdapter((user, self.request), ISerializeToJson)
            return serializer()
        else:
            self.request.response.setStatus(401)
            return



from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from plone.restapi.services import Service
from zope.component import getMultiAdapter


class DeleteWorkingCopy(Service):
    def reply(self):
        policy = ICheckinCheckoutPolicy(self.context)
        working_copy = policy.getWorkingCopy()
        if not policy.getBaseline():
            # We are in the baseline, get the working copy policy
            policy = ICheckinCheckoutPolicy(working_copy)

        control = getMultiAdapter((working_copy, self.request), name="iterate_control")

        if not control.cancel_allowed():
            return self._error(403, "Not authorized", "Cancel not allowed")

        baseline = policy.cancelCheckout()
        baseline.reindexObject()

        return self.reply_no_content()

    def _error(self, status, type, message):
        self.request.response.setStatus(status)
        return {"error": {"type": type, "message": message}}



from plone.dexterity.interfaces import IDexterityContent
from plone.restapi.interfaces import IFieldSerializer
from plone.restapi.interfaces import IJsonCompatible
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.serializer.converters import json_compatible
from plone.restapi.serializer.dxfields import DefaultFieldSerializer
from z3c.relationfield.interfaces import IRelationChoice
from z3c.relationfield.interfaces import IRelationList
from z3c.relationfield.interfaces import IRelationValue
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.globalrequest import getRequest
from zope.interface import implementer
from zope.interface import Interface


@adapter(IRelationValue)
@implementer(IJsonCompatible)
def relationvalue_converter(value):
    if value.to_object:
        summary = getMultiAdapter(
            (value.to_object, getRequest()), ISerializeToJsonSummary
        )()
        return json_compatible(summary)


@adapter(IRelationChoice, IDexterityContent, Interface)
@implementer(IFieldSerializer)
class RelationChoiceFieldSerializer(DefaultFieldSerializer):
    pass


@adapter(IRelationList, IDexterityContent, Interface)
@implementer(IFieldSerializer)
class RelationListFieldSerializer(DefaultFieldSerializer):
    pass



# plone.restapi documentation build configuration file, created by
# sphinx-quickstart on Mon Apr 28 13:04:12 2014.
#
# This file is execfile()d with the current directory set to its containing
# dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath("."))
import os
import sys
import datetime

sys.path.insert(0, os.path.abspath("../../"))

# -- Project information -----------------------------------------------------

# General information about the project.
project = "plone.restapi"
thisyear = datetime.datetime.now().year
copyright = "2014-%s, Plone Foundation" % thisyear

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
# TODO: There must be a way to import this from `setup.py` so we don't have to
# update it manually for each release.
version = "8.21.3.dev0"
release = version

# -- General configuration ----------------------------------------------------

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom ones.
extensions = [
    "sphinxcontrib.httpdomain",
    "sphinxcontrib.httpexample",
]


def patch_pygments_to_highlight_jsonschema():
    """Append "application/json+schema" to the list of mimetypes the
    Pygments JSON lexer is registered to handle.
    """
    try:
        from pygments.lexers._mapping import LEXERS

        mod, lexer_name, aliases, filenames, mimetypes = LEXERS["JsonLexer"]
        mimetypes = mimetypes + ("application/json+schema",)
        LEXERS["JsonLexer"] = (mod, lexer_name, aliases, filenames, mimetypes)
    except:
        # Be defensive (don't fail a docs build if this doesn't work)
        pass


# The suffix of source filenames.
source_suffix = ".rst"
# source_suffix = {
#     ".md": "markdown",
# }


# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ""
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = "%B %d, %Y"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["*ideas/*"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, "()" will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# Options for the linkcheck builder
# Ignore localhost
linkcheck_ignore = [
    r"https://coveralls.io/repos/github/plone/plone.restapi/badge.svg\?branch=master",  # plone.restapi
    r"https://github.com/plone/plone.restapi/blob/dde57b88e0f1b5f5e9f04e6a21865bc0dde55b1c/src/plone/restapi/services/content/add.py#L35-L61",  # plone.restapi
]


# -- Intersphinx configuration ----------------------------------

# This extension can generate automatic links to the documentation of objects
# in other projects. Usage is simple: whenever Sphinx encounters a
# cross-reference that has no matching target in the current documentation set,
# it looks for targets in the documentation sets configured in
# intersphinx_mapping. A reference like :py:class:`zipfile.ZipFile` can then
# linkto the Python documentation for the ZipFile class, without you having to
# specify where it is located exactly.
#
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html
#
intersphinx_mapping = {
    "plone": ("https://6.dev-docs.plone.org/", None),
    "python": ("https://docs.python.org/3/", None),
    "training": ("https://training.plone.org/5/", None),
}


# -- OpenGraph configuration ----------------------------------

ogp_site_url = "https://plonerestapi.readthedocs.org/"
ogp_description_length = 200
ogp_image = "https://docs.plone.org/_static/Plone_logo_square.png"
ogp_site_name = "plone.restapi Documentation"
ogp_type = "website"
ogp_custom_meta_tags = [
    '<meta property="og:locale" content="en_US" />',
]


# -- sphinx_copybutton -----------------------
copybutton_prompt_text = r"^ {0,2}\d{1,3}"
copybutton_prompt_is_regexp = True


# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_book_theme"

html_logo = "_static/logo.svg"
html_favicon = "_static/favicon.ico"

html_css_files = ["custom.css",
                  ("print.css", {"media": "print"})]

# See http://sphinx-doc.org/ext/todo.html#confval-todo_include_todos
todo_include_todos = True

html_theme_options = {
    # TODO: Either get a separate GA ID or enable this one once it is in production.
    # "google_analytics_id": "UA-1907133-6",
    "path_to_docs": "docs",
    "repository_url": "https://github.com/plone/plone.restapi",
    "repository_branch": "master",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
    "extra_navbar": """
    <p class="ploneorglink">
        <a href="https://plone.org">
            <img src="/_static/logo.svg" alt="plone.org" /> plone.org</a>
    </p>""",
    "extra_footer": """<p>The text and illustrations in this website are licensed by the Plone Foundation under a Creative Commons Attribution 4.0 International license. Plone and the Plone® logo are registered trademarks of the Plone Foundation, registered in the United States and other countries. For guidelines on the permitted uses of the Plone trademarks, see <a href="https://plone.org/foundation/logo">https://plone.org/foundation/logo</a>. All other trademarks are owned by their respective owners.</p>
    <p><a href="https://www.netlify.com">
  <img src="https://www.netlify.com/img/global/badges/netlify-color-bg.svg" alt="Deploys by Netlify" />
</a></p>""",
}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = "%(project)s v%(release)s" % {"project": project, "release": release}

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_extra_path = [
    # "robots.txt",
]

# If not "", a "Last updated on:" timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = "%b %d, %Y"

# If true, the Docutils Smart Quotes transform, originally based on SmartyPants
# (limited to English) and currently applying to many languages, will be used
# to convert quotes and dashes to typographically correct entities.
# Note to maintainers: setting this to `True` will cause contractions and
# hyphenated words to be marked as misspelled by spellchecker.
smartquotes = False

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# Announce that we have an opensearch plugin
# https://www.sphinx-doc.org/en/master/usage/configuration.html#confval-html_use_opensearch
html_use_opensearch = "https://plonerestapi.readthedocs.org/"


# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Used by sphinx_sitemap to generate a sitemap
html_baseurl = "https://plonerestapi.readthedocs.org/"


# -- Options for HTML help output -------------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "plonerestapidoc"


# -- Options for myST markdown conversion to html -----------------------------

# For more information see:
# https://myst-parser.readthedocs.io/en/latest/syntax/optional.html
myst_enable_extensions = [
    "deflist",  # You will be able to utilise definition lists
                # https://myst-parser.readthedocs.io/en/latest/syntax/optional.html#definition-lists
    "linkify",  # Identify “bare” web URLs and add hyperlinks.
    "colon_fence",  # You can also use ::: delimiters to denote code fences,\
                    #  instead of ```.
]


# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ("letterpaper" or "a4paper").
    #"papersize": "letterpaper",
    # The font size ("10pt", "11pt" or "12pt").
    #"pointsize": "10pt",
    # Additional stuff for the LaTeX preamble.
    #"preamble": "",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    (
        "index",
        "plonerestapi.tex",
        "plone.restapi Documentation",
        "Plone Foundation",
        "manual",
    ),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ("index", "plonerestapi", "plone.restapi Documentation", ["Plone Foundation"], 1)
]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "plonerestapi",
        "plone.restapi Documentation",
        "Plone Foundation",
        "plonerestapi",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: "footnote", "no", or "inline".
# texinfo_show_urls = "footnote"

suppress_warnings = ["image.nonlocal_uri"]

patch_pygments_to_highlight_jsonschema()
